# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import urllib.parse
import os

import gevent

from volttron.utils import (ClientContext as cc, get_aware_utc_now)

#from volttron.client import get_home
#from volttron.client.agent.utils import get_aware_utc_now
from volttron.client.vip.agent import Agent
from volttron.client import build_vip_address_string

__version__ = "1.0.3"
__author__ = "Craig Allwardt <craig.allwardt@pnnl.gov>"

DEFAULT_TIMEOUT = 30


class Connection(object):
    """A class that manages a connection to a peer and/or server."""

    def __init__(self,
                 address,
                 peer=None,
                 publickey=None,
                 secretkey=None,
                 serverkey=None,
                 volttron_home=None,
                 instance_name=None,
                 message_bus=None,
                 **kwargs):

        self._log = logging.getLogger(__name__)
        self._log.debug("Connection: {}, {}, {}, {}, {}, {}".format(address, peer, publickey,
                                                                    secretkey, serverkey,
                                                                    message_bus))
        self._address = address
        self._peer = peer
        self._serverkey = None
        if peer is None:
            self._log.warning("Peer is non so must be passed in call method.")
        self.volttron_home = volttron_home

        if self.volttron_home is None:
            self.volttron_home = os.path.abspath(cc.get_volttron_home())

        if address.startswith("ipc"):
            full_address = address
        else:
            parsed = urllib.parse.urlparse(address)
            if parsed.scheme == "tcp":
                qs = urllib.parse.parse_qs(parsed.query)
                self._log.debug("QS IS: {}".format(qs))
                if "serverkey" in qs:
                    self._serverkey = qs.get("serverkey")
                else:
                    self._serverkey = serverkey

                # Handle case when the address has all the information in it.
                if "serverkey" in qs and "publickey" in qs and "secretkey" in qs:
                    full_address = address
                else:
                    full_address = build_vip_address_string(
                        vip_root=address,
                        serverkey=serverkey,
                        publickey=publickey,
                        secretkey=secretkey,
                    )
            elif parsed.scheme == "ipc":
                full_address = address
            else:
                raise AttributeError("Invalid address type specified. ipc or tcp accepted.")

        self._server = Agent(address=full_address,
                             volttron_home=self.volttron_home,
                             enable_store=False,
                             reconnect_interval=1000,
                             instance_name=instance_name,
                             message_bus=message_bus,
                             **kwargs)
        # TODO the following should work as well, but doesn't.  Not sure why!
        # self._server = Agent(address=address, serverkey=serverkey,
        #                      secretkey=secretkey, publickey=publickey,
        #                      enable_store=False,
        #                      volttron_home=self.volttron_home, **kwargs)
        self._greenlet = None
        self._connected_since = None
        self._last_publish = None
        self._last_publish_failed = False
        self._last_rpc_call = None

        # Make the actual attempt to connect to the server.
        self.is_connected()

    @property
    def serverkey(self):
        return self._serverkey

    @property
    def last_publish_failed(self):
        return self._last_publish_failed

    @property
    def connected_since(self):
        return self._connected_since

    @property
    def last_publish(self):
        return self._last_publish

    @property
    def last_rpc_call(self):
        return self._last_rpc_call

    @property
    def address(self):
        return self._address

    @property
    def peer(self):
        return self._peer

    @property
    def server(self):
        if self._greenlet is None:
            event = gevent.event.Event()
            self._greenlet = gevent.spawn(self._server.core.run, event)

            try:
                with gevent.Timeout(DEFAULT_TIMEOUT):
                    event.wait()
            except gevent.Timeout:
                self.kill()
                self._greenlet = None
                raise

            self._connected_since = get_aware_utc_now()
            if self.peer:
                if self.peer not in self._server.vip.peerlist().get(timeout=2):
                    self._log.warning("peer {} not found connected to router.".format(self.peer))
        return self._server

    def peers(self, timeout=DEFAULT_TIMEOUT):
        return self.server.vip.peerlist().get(timeout=timeout)

    def is_connected(self, timeout=DEFAULT_TIMEOUT):
        return self.server.core.connected and self.is_peer_connected(timeout)

    def is_peer_connected(self, timeout=DEFAULT_TIMEOUT):
        self._log.debug("Checking for peer {}".format(self.peer))
        return self.peer in self.peers()

    def publish(self, topic, headers=None, message=None, timeout=DEFAULT_TIMEOUT):
        if timeout is None:
            raise ValueError("timeout cannot be None")

        timeout = int(timeout)

        self.server.vip.pubsub.publish("pubsub", topic=topic, headers=headers,
                                       message=message).get(timeout=timeout)

    def subscribe(self, prefix, callback):
        self.server.vip.pubsub.subscribe("pubsub", prefix, callback)

    def call(self, method, *args, **kwargs):
        timeout = kwargs.pop("timeout", DEFAULT_TIMEOUT)
        peer = kwargs.pop("peer", None)

        if peer is not None:
            return self.server.vip.rpc.call(peer, method, *args, **kwargs).get(timeout=timeout)

        if self.peer is not None:
            return self.server.vip.rpc.call(self.peer, method, *args,
                                            **kwargs).get(timeout=timeout)

        raise ValueError("peer not specified on class or as method argument.")

    def notify(self, method, *args, **kwargs):
        peer = kwargs.pop("peer")

        if peer is not None:
            return self.server.vip.rpc.notify(peer, method, *args, **kwargs)

        if self.peer is not None:
            return self.server.vip.rpc.notify(self.peer, method, *args, **kwargs)

        raise ValueError("peer not specified on class or as method argument.")

    def kill(self, *args, **kwargs):
        if self._greenlet is not None:
            self._greenlet.kill(*args, **kwargs)
            del self._greenlet
            self._greenlet = None



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging

_log = logging.getLogger(__name__)


class ServicePeerNotifier(object):
    """
    This class is responsible for routing the base_router's connections and disconnections
    from the zmq thread through to the registered callback functions.
    """

    def __init__(self):
        self._registered_added = set()
        self._registered_dropped = set()

    def register_peer_callback(self, added_callback, dropped_callback):
        """
        Register functions for adding callbacks for connected and disconnected peers
        to the message bus.

        The signature of the callback should be:

        .. code-block :: python

            def added_callback(peer):
                # the peer is a string identity connected.
                pass

        :param added_callback:
        :param dropped_callback:
        """
        assert added_callback is not None
        assert dropped_callback is not None

        self._registered_added.add(added_callback)
        self._registered_dropped.add(dropped_callback)

    def peer_added(self, peer):
        """
        Handles calling registered methods
        :param peer:
        :return:
        """
        for fn in self._registered_added:
            fn(peer)

    def peer_dropped(self, peer):
        """
        Handles calling of registered methods when a peer drops a connection to the platform.
        :param peer:
        :return:
        """
        for fn in self._registered_dropped:
            fn(peer)


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from __future__ import annotations

__all__ = [
    "AuthService", "AuthFile", "AuthEntry", "AuthFileEntryAlreadyExists", "AuthFileIndexError",
    "AuthException"
]

import bisect
import logging
import os
import random
import re
import shutil
from typing import Optional
import uuid
from collections import defaultdict

import gevent
import gevent.core
from gevent.fileobject import FileObject
from zmq import green as zmq

from volttron.types import ServiceInterface
from volttron.utils import (
    ClientContext as cc,
    create_file_if_missing,
    strip_comments,
)
from volttron.utils import jsonapi
from volttron.utils.filewatch import watch_file
from volttron.utils.certs import Certs
from volttron.utils.keystore import encode_key, BASE64_ENCODED_CURVE_KEY_LEN
from volttron.client.vip.agent import Agent, Core, RPC, VIPError
from volttron.client.known_identities import (
    VOLTTRON_CENTRAL_PLATFORM,
    CONTROL,
    CONTROL_CONNECTION,
)

# TODO: it seems this should not be so nested of a import path.
from volttron.client.vip.agent.subsystems.pubsub import ProtectedPubSubTopics
import volttron.types.server_config as server_config

# from volttron.platform.certs import Certs
# from volttron.platform.vip.agent.errors import VIPError
# from volttron.platform.vip.pubsubservice import ProtectedPubSubTopics
# from .agent.utils import strip_comments, create_file_if_missing, watch_file, get_messagebus
# from .vip.agent import Agent, Core, RPC
# from .vip.socket import encode_key, BASE64_ENCODED_CURVE_KEY_LEN

_log = logging.getLogger(__name__)

_dump_re = re.compile(r"([,\\])")
_load_re = re.compile(r"\\(.)|,")


def isregex(obj):
    return len(obj) > 1 and obj[0] == obj[-1] == "/"


def dump_user(*args):
    return ",".join([_dump_re.sub(r"\\\1", arg) for arg in args])


def load_user(string):

    def sub(match):
        return match.group(1) or "\x00"

    return _load_re.sub(sub, string).split("\x00")


class AuthException(Exception):
    """General exception for any auth error"""

    pass


class AuthService(ServiceInterface):

    def __init__(self, server_config, **kwargs):
        #auth_file, protected_topics_file, setup_mode, aip, *args, **kwargs):
        self.allow_any = kwargs.pop("allow_any", False)

        super(AuthService, self).__init__(**kwargs)

        self._server_config = server_config
        # This agent is started before the router so we need
        # to keep it from blocking.
        self.core.delay_running_event_set = False
        self._certs = None
        if cc.get_messagebus() == "rmq":
            self._certs = Certs()
        self.auth_file_path = str(self._server_config.auth_file)
        self.auth_file = AuthFile(self.auth_file_path)
        self.aip = self._server_config.aip
        self.zap_socket = None
        self._zap_greenlet = None
        self.auth_entries = []
        self._is_connected = False
        self._protected_topics_file_path = str(self._server_config.protected_topics_file)
        self._protected_topics_file = str(self._server_config.protected_topics_file)
        self._protected_topics_for_rmq = ProtectedPubSubTopics()
        self._setup_mode = self._server_config.opts.setup_mode
        self._auth_pending = []
        self._auth_denied = []
        self._auth_approved = []

        def topics():
            return defaultdict(set)

        self._user_to_permissions = topics()

    @Core.receiver("onsetup")
    def setup_zap(self, sender, **kwargs):
        self.zap_socket = zmq.Socket(zmq.Context.instance(), zmq.ROUTER)
        self.zap_socket.bind("inproc://zeromq.zap.01")
        if self.allow_any:
            _log.warning("insecure permissive authentication enabled")
        self.read_auth_file()
        self._read_protected_topics_file()
        self.core.spawn(watch_file, self.auth_file_path, self.read_auth_file)
        self.core.spawn(
            watch_file,
            self._protected_topics_file_path,
            self._read_protected_topics_file,
        )
        if self.core.messagebus == "rmq":
            self.vip.peerlist.onadd.connect(self._check_topic_rules)

    def _update_auth_lists(self, entries, is_allow=True):
        auth_list = []
        for entry in entries:
            auth_list.append({
                "domain": entry.domain,
                "address": entry.address,
                "mechanism": entry.mechanism,
                "credentials": entry.credentials,
                "user_id": entry.user_id,
                "retries": 0,
            })
        if is_allow:
            self._auth_approved = [entry for entry in auth_list if entry["address"] is not None]
        else:
            self._auth_denied = [entry for entry in auth_list if entry["address"] is not None]

    def read_auth_file(self):
        _log.info("loading auth file %s", self.auth_file_path)
        entries = self.auth_file.read_allow_entries()
        denied_entries = self.auth_file.read_deny_entries()
        # Populate auth lists with current entries
        self._update_auth_lists(entries)
        self._update_auth_lists(denied_entries, is_allow=False)

        entries = [entry for entry in entries if entry.enabled]
        # sort the entries so the regex credentails follow the concrete creds
        entries.sort()
        self.auth_entries = entries
        if self._is_connected:
            try:
                _log.debug("Sending auth updates to peers")
                # Give it few seconds for platform to startup or for the
                # router to detect agent install/remove action
                gevent.sleep(2)
                self._send_update()
            except BaseException as e:
                _log.error("Exception sending auth updates to peer. {}".format(e))
                raise e
        _log.info("auth file %s loaded", self.auth_file_path)

    def get_protected_topics(self):
        protected = self._protected_topics
        return protected

    def _read_protected_topics_file(self):
        # Read protected topics file and send to router
        try:
            create_file_if_missing(self._protected_topics_file)
            with open(self._protected_topics_file) as fil:
                # Use gevent FileObject to avoid blocking the thread
                data = FileObject(fil, close=False).read()
                self._protected_topics = jsonapi.loads(data) if data else {}
                if self.core.messagebus == "rmq":
                    self._load_protected_topics_for_rmq()
                    # Deferring the RMQ topic permissions to after "onstart" event
                else:
                    self._send_protected_update_to_pubsub(self._protected_topics)
        except Exception:
            _log.exception("error loading %s", self._protected_topics_file)

    def _send_update(self):
        user_to_caps = self.get_user_to_capabilities()
        i = 0
        exception = None
        peers = None
        # peerlist times out lots of times when running test suite. This happens even with higher timeout in get()
        # but if we retry peerlist succeeds by second attempt most of the time!!!
        while not peers and i < 3:
            try:
                i = i + 1
                peers = self.vip.peerlist().get(timeout=0.5)
            except BaseException as e:
                _log.warning("Attempt {} to get peerlist failed with exception {}".format(i, e))
                peers = list(self.vip.peerlist.peers_list)
                _log.warning("Get list of peers from subsystem directly".format(peers))
                exception = e

        if not peers:
            raise BaseException("No peers connected to the platform")

        _log.debug("after getting peerlist to send auth updates")

        for peer in peers:
            if peer not in [self.core.identity, CONTROL_CONNECTION]:
                _log.debug(f"Sending auth update to peers {peer}")
                self.vip.rpc.call(peer, "auth.update", user_to_caps)
        if self.core.messagebus == "rmq":
            self._check_rmq_topic_permissions()
        else:
            self._send_auth_update_to_pubsub()

    def _send_auth_update_to_pubsub(self):
        user_to_caps = self.get_user_to_capabilities()
        # Send auth update message to router
        json_msg = jsonapi.dumpb(dict(capabilities=user_to_caps))
        frames = [zmq.Frame(b"auth_update"), zmq.Frame(json_msg)]
        # <recipient, subsystem, args, msg_id, flags>
        self.core.socket.send_vip(b"", b"pubsub", frames, copy=False)

    def _send_protected_update_to_pubsub(self, contents):
        protected_topics_msg = jsonapi.dumpb(contents)

        frames = [zmq.Frame(b"protected_update"), zmq.Frame(protected_topics_msg)]
        if self._is_connected:
            try:
                # <recipient, subsystem, args, msg_id, flags>
                self.core.socket.send_vip(b"", b"pubsub", frames, copy=False)
            except VIPError as ex:
                _log.error("Error in sending protected topics update to clear PubSub: " + str(ex))

    @Core.receiver("onstop")
    def stop_zap(self, sender, **kwargs):
        if self._zap_greenlet is not None:
            self._zap_greenlet.kill()

    @Core.receiver("onfinish")
    def unbind_zap(self, sender, **kwargs):
        if self.zap_socket is not None:
            self.zap_socket.unbind("inproc://zeromq.zap.01")

    @Core.receiver("onstart")
    def zap_loop(self, sender, **kwargs):
        """
        The zap loop is the starting of the authentication process for
        the VOLTTRON zmq message bus.  It talks directly with the low
        level socket so all responses must be byte like objects, in
        this case we are going to send zmq frames across the wire.

        :param sender:
        :param kwargs:
        :return:
        """
        self._is_connected = True
        self._zap_greenlet = gevent.getcurrent()
        sock = self.zap_socket
        time = gevent.core.time
        blocked = {}
        wait_list = []
        timeout = None
        if self.core.messagebus == "rmq":
            # Check the topic permissions of all the connected agents
            self._check_rmq_topic_permissions()
        else:
            self._send_protected_update_to_pubsub(self._protected_topics)

        while True:
            events = sock.poll(timeout)
            now = time()
            if events:
                zap = sock.recv_multipart()

                version = zap[2]
                if version != b"1.0":
                    continue
                domain, address, userid, kind = zap[4:8]
                credentials = zap[8:]
                if kind == b"CURVE":
                    credentials[0] = encode_key(credentials[0])
                elif kind not in [b"NULL", b"PLAIN"]:
                    continue
                response = zap[:4]
                domain = domain.decode("utf-8")
                address = address.decode("utf-8")
                kind = kind.decode("utf-8")
                user = self.authenticate(domain, address, kind, credentials)
                _log.info(f"AUTH: After authenticate user id: {user}, {userid}")
                if user:
                    _log.info(
                        "authentication success: userid=%r domain=%r, address=%r, "
                        "mechanism=%r, credentials=%r, user=%r",
                        userid,
                        domain,
                        address,
                        kind,
                        credentials[:1],
                        user,
                    )
                    response.extend([b"200", b"SUCCESS", user.encode("utf-8"), b""])
                    sock.send_multipart(response)
                else:
                    userid = str(uuid.uuid4())
                    _log.info(
                        "authentication failure: userid=%r, domain=%r, address=%r, "
                        "mechanism=%r, credentials=%r",
                        userid,
                        domain,
                        address,
                        kind,
                        credentials,
                    )
                    # If in setup mode, add/update auth entry
                    if self._setup_mode:
                        self._update_auth_entry(domain, address, kind, credentials[0], userid)
                        _log.info(
                            "new authentication entry added in setup mode: domain=%r, address=%r, "
                            "mechanism=%r, credentials=%r, user_id=%r",
                            domain,
                            address,
                            kind,
                            credentials[:1],
                            userid,
                        )
                        response.extend([b"200", b"SUCCESS", b"", b""])
                        _log.debug("AUTH response: {}".format(response))
                        sock.send_multipart(response)
                    else:
                        if type(userid) == bytes:
                            userid = userid.decode("utf-8")
                        self._update_auth_pending(domain, address, kind, credentials[0], userid)

                    try:
                        expire, delay = blocked[address]
                    except KeyError:
                        delay = random.random()
                    else:
                        if now >= expire:
                            delay = random.random()
                        else:
                            delay *= 2
                            if delay > 100:
                                delay = 100
                    expire = now + delay
                    bisect.bisect(wait_list, (expire, address, response))
                    blocked[address] = expire, delay
            while wait_list:
                expire, address, response = wait_list[0]
                if now < expire:
                    break
                wait_list.pop(0)
                response.extend([b"400", b"FAIL", b"", b""])
                sock.send_multipart(response)
                try:
                    if now >= blocked[address][0]:
                        blocked.pop(address)
                except KeyError:
                    pass
            timeout = (wait_list[0][0] - now) if wait_list else None

    def authenticate(self, domain, address, mechanism, credentials):
        for entry in self.auth_entries:
            if entry.match(domain, address, mechanism, credentials):
                return entry.user_id or dump_user(domain, address, mechanism, *credentials[:1])
        if mechanism == "NULL" and address.startswith("localhost:"):
            parts = address.split(":")[1:]
            if len(parts) > 2:
                pid = int(parts[2])
                agent_uuid = self.aip.agent_uuid_from_pid(pid)
                if agent_uuid:
                    return dump_user(domain, address, "AGENT", agent_uuid)
            uid = int(parts[0])
            if uid == os.getuid():
                return dump_user(domain, address, mechanism, *credentials[:1])
        if self.allow_any:
            return dump_user(domain, address, mechanism, *credentials[:1])

    @RPC.export
    def get_user_to_capabilities(self):
        """RPC method

        Gets a mapping of all users to their capabiliites.

        :returns: mapping of users to capabilities
        :rtype: dict
        """
        user_to_caps = {}
        for entry in self.auth_entries:
            user_to_caps[entry.user_id] = entry.capabilities
        return user_to_caps

    @RPC.export
    def get_authorizations(self, user_id):
        """RPC method

        Gets capabilities, groups, and roles for a given user.

        :param user_id: user id field from VOLTTRON Interconnect Protocol
        :type user_id: str
        :returns: tuple of capabiliy-list, group-list, role-list
        :rtype: tuple
        """
        use_parts = True
        try:
            domain, address, mechanism, credentials = load_user(user_id)
        except ValueError:
            use_parts = False
        for entry in self.auth_entries:
            if entry.user_id == user_id:
                return [entry.capabilities, entry.groups, entry.roles]
            elif use_parts:
                if entry.match(domain, address, mechanism, [credentials]):
                    return entry.capabilities, entry.groups, entry.roles

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def approve_authorization_failure(self, user_id):
        """RPC method

        Approves a pending CSR or credential, based on provided identity.
        The approved CSR or credential can be deleted or denied later.
        An approved credential is stored in the allow list in auth.json.

        :param user_id: user id field from VOLTTRON Interconnect Protocol or common name for CSR
        :type user_id: str
        """

        val_err = None
        if self._certs:
            # Will fail with ValueError when a zmq credential user_id is passed.
            try:
                self._certs.approve_csr(user_id)
                permissions = self.core.rmq_mgmt.get_default_permissions(user_id)

                if (
                        "federation" in user_id
                ):    # federation needs more than the current default permissions # TODO: Fix authorization in rabbitmq
                    permissions = dict(configure=".*", read=".*", write=".*")
                self.core.rmq_mgmt.create_user_with_permissions(user_id, permissions, True)
                _log.debug("Created cert and permissions for user: {}".format(user_id))
            # Stores error message in case it is caused by an unexpected failure
            except ValueError as e:
                val_err = e
        index = 0
        matched_index = -1
        for pending in self._auth_pending:
            if user_id == pending["user_id"]:
                self._update_auth_entry(
                    pending["domain"],
                    pending["address"],
                    pending["mechanism"],
                    pending["credentials"],
                    pending["user_id"],
                )
                matched_index = index
                val_err = None
                break
            index = index + 1
        if matched_index >= 0:
            del self._auth_pending[matched_index]

        for pending in self._auth_denied:
            if user_id == pending["user_id"]:
                self._update_auth_entry(
                    pending["domain"],
                    pending["address"],
                    pending["mechanism"],
                    pending["credentials"],
                    pending["user_id"],
                )
                self._remove_auth_entry(pending["credentials"], is_allow=False)
                val_err = None
        # If the user_id supplied was not for a ZMQ credential, and the pending_csr check failed,
        # output the ValueError message to the error log.
        if val_err:
            _log.error(f"{val_err}")

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def deny_authorization_failure(self, user_id):
        """RPC method

        Denies a pending CSR or credential, based on provided identity.
        The denied CSR or credential can be deleted or accepted later.
        A denied credential is stored in the deny list in auth.json.

        :param user_id: user id field from VOLTTRON Interconnect Protocol or common name for CSR
        :type user_id: str
        """

        val_err = None
        if self._certs:
            # Will fail with ValueError when a zmq credential user_id is passed.
            try:
                self._certs.deny_csr(user_id)
                _log.debug("Denied cert for user: {}".format(user_id))
            # Stores error message in case it is caused by an unexpected failure
            except ValueError as e:
                val_err = e

        index = 0
        matched_index = -1
        for pending in self._auth_pending:
            if user_id == pending["user_id"]:
                self._update_auth_entry(
                    pending["domain"],
                    pending["address"],
                    pending["mechanism"],
                    pending["credentials"],
                    pending["user_id"],
                    is_allow=False,
                )
                matched_index = index
                val_err = None
                break
            index = index + 1
        if matched_index >= 0:
            del self._auth_pending[matched_index]

        for pending in self._auth_approved:
            if user_id == pending["user_id"]:
                self._update_auth_entry(
                    pending["domain"],
                    pending["address"],
                    pending["mechanism"],
                    pending["credentials"],
                    pending["user_id"],
                    is_allow=False,
                )
                self._remove_auth_entry(pending["credentials"])
                val_err = None
        # If the user_id supplied was not for a ZMQ credential, and the pending_csr check failed,
        # output the ValueError message to the error log.
        if val_err:
            _log.error(f"{val_err}")

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def delete_authorization_failure(self, user_id):
        """RPC method

        Deletes a pending CSR or credential, based on provided identity.
        To approve or deny a deleted pending CSR or credential,
        the request must be resent by the remote platform or agent.

        :param user_id: user id field from VOLTTRON Interconnect Protocol or common name for CSR
        :type user_id: str
        """

        val_err = None
        if self._certs:
            # Will fail with ValueError when a zmq credential user_id is passed.
            try:
                self._certs.delete_csr(user_id)
                _log.debug("Denied cert for user: {}".format(user_id))
            # Stores error message in case it is caused by an unexpected failure
            except ValueError as e:
                val_err = e

        index = 0
        matched_index = -1
        for pending in self._auth_pending:
            if user_id == pending["user_id"]:
                self._update_auth_entry(
                    pending["domain"],
                    pending["address"],
                    pending["mechanism"],
                    pending["credentials"],
                    pending["user_id"],
                )
                matched_index = index
                val_err = None
                break
            index = index + 1
        if matched_index >= 0:
            del self._auth_pending[matched_index]

        index = 0
        matched_index = -1
        for pending in self._auth_pending:
            if user_id == pending["user_id"]:
                matched_index = index
                val_err = None
                break
            index = index + 1
        if matched_index >= 0:
            del self._auth_pending[matched_index]

        for pending in self._auth_approved:
            if user_id == pending["user_id"]:
                self._remove_auth_entry(pending["credentials"])
                val_err = None

        for pending in self._auth_denied:
            if user_id == pending["user_id"]:
                self._remove_auth_entry(pending["credentials"], is_allow=False)
                val_err = None

        # If the user_id supplied was not for a ZMQ credential, and the pending_csr check failed,
        # output the ValueError message to the error log.
        if val_err:
            _log.error(f"{val_err}")

    @RPC.export
    def get_authorization_pending(self):
        """RPC method

        Returns a list of failed (pending) ZMQ credentials.

        :rtype: list
        """
        return list(self._auth_pending)

    @RPC.export
    def get_authorization_approved(self):
        """RPC method

        Returns a list of approved ZMQ credentials.
        This list is updated whenever the auth file is read.
        It includes all allow entries from the auth file that contain a populated address field.

        :rtype: list
        """
        return list(self._auth_approved)

    @RPC.export
    def get_authorization_denied(self):
        """RPC method

        Returns a list of denied ZMQ credentials.
        This list is updated whenever the auth file is read.
        It includes all deny entries from the auth file that contain a populated address field.

        :rtype: list
        """
        return list(self._auth_denied)

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def get_pending_csrs(self):
        """RPC method

        Returns a list of pending CSRs.
        This method provides RPC access to the Certs class's get_pending_csr_requests method.
        This method is only applicable for web-enabled, RMQ instances.

        :rtype: list
        """
        if self._certs:
            csrs = [c for c in self._certs.get_pending_csr_requests()]
            return csrs
        else:
            return []

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def get_pending_csr_status(self, common_name):
        """RPC method

        Returns the status of a pending CSRs.
        This method provides RPC access to the Certs class's get_csr_status method.
        This method is only applicable for web-enabled, RMQ instances.
        Currently, this method is only used by admin_endpoints.

        :param common_name: Common name for CSR
        :type common_name: str
        :rtype: str
        """
        if self._certs:
            return self._certs.get_csr_status(common_name)
        else:
            return ""

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def get_pending_csr_cert(self, common_name):
        """RPC method

        Returns the cert of a pending CSRs.
        This method provides RPC access to the Certs class's get_cert_from_csr method.
        This method is only applicable for web-enabled, RMQ instances.
        Currently, this method is only used by admin_endpoints.

        :param common_name: Common name for CSR
        :type common_name: str
        :rtype: str
        """
        if self._certs:
            return self._certs.get_cert_from_csr(common_name).decode("utf-8")
        else:
            return ""

    @RPC.export
    @RPC.allow(capabilities="allow_auth_modifications")
    def get_all_pending_csr_subjects(self):
        """RPC method

        Returns a list of all certs subjects.
        This method provides RPC access to the Certs class's get_all_cert_subjects method.
        This method is only applicable for web-enabled, RMQ instances.
        Currently, this method is only used by admin_endpoints.

        :rtype: list
        """
        if self._certs:
            return self._certs.get_all_cert_subjects()
        else:
            return []

    def _get_authorizations(self, user_id, index):
        """Convenience method for getting authorization component by index"""
        auths = self.get_authorizations(user_id)
        if auths:
            return auths[index]
        return []

    @RPC.export
    def get_capabilities(self, user_id):
        """RPC method

        Gets capabilities for a given user.

        :param user_id: user id field from VOLTTRON Interconnect Protocol
        :type user_id: str
        :returns: list of capabilities
        :rtype: list
        """
        return self._get_authorizations(user_id, 0)

    @RPC.export
    def get_groups(self, user_id):
        """RPC method

        Gets groups for a given user.

        :param user_id: user id field from VOLTTRON Interconnect Protocol
        :type user_id: str
        :returns: list of groups
        :rtype: list
        """
        return self._get_authorizations(user_id, 1)

    @RPC.export
    def get_roles(self, user_id):
        """RPC method

        Gets roles for a given user.

        :param user_id: user id field from VOLTTRON Interconnect Protocol
        :type user_id: str
        :returns: list of roles
        :rtype: list
        """
        return self._get_authorizations(user_id, 2)

    def _update_auth_entry(self, domain, address, mechanism, credential, user_id, is_allow=True):
        # Make a new entry
        fields = {
            "domain": domain,
            "address": address,
            "mechanism": mechanism,
            "credentials": credential,
            "user_id": user_id,
            "groups": "",
            "roles": "",
            "capabilities": "",
            "comments": "Auth entry added in setup mode",
        }
        new_entry = AuthEntry(**fields)

        try:
            self.auth_file.add(new_entry, overwrite=False, is_allow=is_allow)
        except AuthException as err:
            _log.error("ERROR: %s\n" % str(err))

    def _remove_auth_entry(self, credential, is_allow=True):
        try:
            self.auth_file.remove_by_credentials(credential, is_allow=is_allow)
        except AuthException as err:
            _log.error("ERROR: %s\n" % str(err))

    def _update_auth_pending(self, domain, address, mechanism, credential, user_id):
        for entry in self._auth_denied:
            # Check if failure entry has been denied. If so, increment the failure's denied count
            if ((entry["domain"] == domain) and (entry["address"] == address)
                    and (entry["mechanism"] == mechanism)
                    and (entry["credentials"] == credential)):
                entry["retries"] += 1
                return

        for entry in self._auth_pending:
            # Check if failure entry exists. If so, increment the failure count
            if ((entry["domain"] == domain) and (entry["address"] == address)
                    and (entry["mechanism"] == mechanism)
                    and (entry["credentials"] == credential)):
                entry["retries"] += 1
                return
        # Add a new failure entry
        fields = {
            "domain": domain,
            "address": address,
            "mechanism": mechanism,
            "credentials": credential,
            "user_id": user_id,
            "retries": 1,
        }
        self._auth_pending.append(dict(fields))
        return

    def _load_protected_topics_for_rmq(self):
        try:
            write_protect = self._protected_topics["write-protect"]
        except KeyError:
            write_protect = []

        topics = ProtectedPubSubTopics()
        try:
            for entry in write_protect:
                topics.add(entry["topic"], entry["capabilities"])
        except KeyError:
            _log.exception("invalid format for protected topics ")
        else:
            self._protected_topics_for_rmq = topics

    def _check_topic_rules(self, sender, **kwargs):
        delay = 0.05
        self.core.spawn_later(delay, self._check_rmq_topic_permissions)

    def _check_rmq_topic_permissions(self):
        """
        Go through the topic permissions for each agent based on the protected topic setting.
        Update the permissions for the agent/user based on the latest configuration
        :return:
        """
        return
        # Get agent to capabilities mapping
        user_to_caps = self.get_user_to_capabilities()
        # Get topics to capabilities mapping
        topic_to_caps = self._protected_topics_for_rmq.get_topic_caps()    # topic to caps

        peers = self.vip.peerlist().get(timeout=5)
        # _log.debug("USER TO CAPS: {0}, TOPICS TO CAPS: {1}, {2}".format(user_to_caps,
        #                                                                 topic_to_caps,
        #                                                                 self._user_to_permissions))
        if not user_to_caps or not topic_to_caps:
            # clear all old permission rules
            for peer in peers:
                self._user_to_permissions[peer].clear()
        else:
            for topic, caps_for_topic in topic_to_caps.items():
                for user in user_to_caps:
                    try:
                        caps_for_user = user_to_caps[user]
                        common_caps = list(set(caps_for_user).intersection(caps_for_topic))
                        if common_caps:
                            self._user_to_permissions[user].add(topic)
                        else:
                            try:
                                self._user_to_permissions[user].remove(topic)
                            except KeyError as e:
                                if not self._user_to_permissions[user]:
                                    self._user_to_permissions[user] = set()
                    except KeyError as e:
                        try:
                            self._user_to_permissions[user].remove(topic)
                        except KeyError as e:
                            if not self._user_to_permissions[user]:
                                self._user_to_permissions[user] = set()

        all = set()
        for user in user_to_caps:
            all.update(self._user_to_permissions[user])

        # Set topic permissions now
        for peer in peers:
            not_allowed = all.difference(self._user_to_permissions[peer])
            self._update_topic_permission_tokens(peer, not_allowed)

    def _update_topic_permission_tokens(self, identity, not_allowed):
        """
        Make rules for read and write permission on topic (routing key)
        for an agent based on protected topics setting
        :param identity: identity of the agent
        :return:
        """
        read_tokens = [
            "{instance}.{identity}".format(instance=self.core.instance_name, identity=identity),
            "__pubsub__.*",
        ]
        write_tokens = ["{instance}.*".format(instance=self.core.instance_name, identity=identity)]

        if not not_allowed:
            write_tokens.append("__pubsub__.{instance}.*".format(instance=self.core.instance_name))
        else:
            not_allowed_string = "|".join(not_allowed)
            write_tokens.append("__pubsub__.{instance}.".format(instance=self.core.instance_name) +
                                "^(!({not_allow})).*$".format(not_allow=not_allowed_string))
        current = self.core.rmq_mgmt.get_topic_permissions_for_user(identity)
        # _log.debug("CURRENT for identity: {0}, {1}".format(identity, current))
        if current and isinstance(current, list):
            current = current[0]
            dift = False
            read_allowed_str = "|".join(read_tokens)
            write_allowed_str = "|".join(write_tokens)
            if re.search(current["read"], read_allowed_str):
                dift = True
                current["read"] = read_allowed_str
            if re.search(current["write"], write_allowed_str):
                dift = True
                current["write"] = write_allowed_str
                # _log.debug("NEW {0}, DIFF: {1} ".format(current, dift))
                # if dift:
                #     set_topic_permissions_for_user(current, identity)
        else:
            current = dict()
            current["exchange"] = "volttron"
            current["read"] = "|".join(read_tokens)
            current["write"] = "|".join(write_tokens)
            # _log.debug("NEW {0}, New string ".format(current))
            # set_topic_permissions_for_user(current, identity)

    def _check_token(self, actual, allowed):
        pending = actual[:]
        for tk in actual:
            if tk in allowed:
                pending.remove(tk)
        return pending


class String(str):

    def __new__(cls, value):
        obj = super(String, cls).__new__(cls, value)
        if isregex(obj):
            obj.regex = regex = re.compile("^" + obj[1:-1] + "$")
            obj.match = lambda val: bool(regex.match(val))
        return obj

    def match(self, value):
        return value == self


class List(list):

    def match(self, value):
        for elem in self:
            if elem.match(value):
                return True
        return False


class AuthEntryInvalid(AuthException):
    """Exception for invalid AuthEntry objects"""

    pass


class AuthEntry(object):
    """An authentication entry contains fields for authenticating and
    granting permissions to an agent that connects to the platform.

    :param str domain: Name assigned to locally bound address
    :param str address: Remote address of the agent
    :param str mechanism: Authentication mechanism, valid options are
        'NULL' (no authentication), 'PLAIN' (username/password),
        'CURVE' (CurveMQ public/private keys)
    :param str credentials: Value depends on `mechanism` parameter:
        `None` if mechanism is 'NULL'; password if mechanism is
        'PLAIN'; encoded public key if mechanism is 'CURVE' (see
        :py:meth:`volttron.platform.vip.socket.encode_key` for method
        to encode public key)
    :param str user_id: Name to associate with agent (Note: this does
        not have to match the agent's VIP identity)
    :param list capabilities: Authorized capabilities for this agent
    :param list roles: Authorized roles for this agent. (Role names map
        to a set of capabilities)
    :param list groups: Authorized groups for this agent. (Group names
        map to a set of roles)
    :param str comments: Comments to associate with entry
    :param bool enabled: Entry will only be used if this value is True
    :param kwargs: These extra arguments will be ignored
    """

    def __init__(
        self,
        domain=None,
        address=None,
        mechanism="CURVE",
        credentials=None,
        user_id=None,
        groups=None,
        roles=None,
        capabilities: Optional[dict] = None,
        comments=None,
        enabled=True,
        **kwargs,
    ):

        self.domain = AuthEntry._build_field(domain)
        self.address = AuthEntry._build_field(address)
        self.mechanism = mechanism
        self.credentials = AuthEntry._build_field(credentials)
        self.groups = AuthEntry._build_field(groups) or []
        self.roles = AuthEntry._build_field(roles) or []
        self.capabilities = AuthEntry.build_capabilities_field(capabilities) or {}
        self.comments = AuthEntry._build_field(comments)
        if user_id is None:
            user_id = str(uuid.uuid4())
        self.user_id = user_id
        self.enabled = enabled
        if kwargs:
            _log.debug("auth record has unrecognized keys: %r" % (list(kwargs.keys()), ))
        self._check_validity()

    def __lt__(self, other):
        """Entries with non-regex credentials will be less than regex
        credentials. When sorted, the non-regex credentials will be
        checked first."""
        try:
            self.credentials.regex
        except AttributeError:
            return True
        return False

    @staticmethod
    def _build_field(value):
        if not value:
            return None
        if isinstance(value, str):
            return String(value)
        return List(String(elem) for elem in value)

    @staticmethod
    def build_capabilities_field(value: Optional[dict]):
        # _log.debug("_build_capabilities {}".format(value))

        if not value:
            return None

        if isinstance(value, list):
            result = dict()
            for elem in value:
                # update if it is not there or if existing entry doesn't have args.
                # i.e. capability with args can override capability str
                temp = result.update(AuthEntry._get_capability(elem))
                if temp and result[next(iter(temp))] is None:
                    result.update(temp)
            _log.debug("Returning field _build_capabilities {}".format(result))
            return result
        else:
            return AuthEntry._get_capability(value)

    @staticmethod
    def _get_capability(value):
        err_message = (
            "Invalid capability value: {} of type {}. Capability entries can only be a string or "
            "dictionary or list containing string/dictionary. "
            "dictionaries should be of the format {'capability_name':None} or "
            "{'capability_name':{'arg1':'value',...}")
        if isinstance(value, str):
            return {value: None}
        elif isinstance(value, dict):
            return value
        else:
            raise AuthEntryInvalid(err_message.format(value, type(value)))

    def add_capabilities(self, capabilities):
        temp = AuthEntry.build_capabilities_field(capabilities)
        if temp:
            self.capabilities.update(temp)

    def match(self, domain, address, mechanism, credentials):
        return ((self.domain is None or self.domain.match(domain))
                and (self.address is None or self.address.match(address))
                and self.mechanism == mechanism
                and (self.mechanism == "NULL" or
                     (len(self.credentials) > 0 and self.credentials.match(credentials[0]))))

    def __str__(self):
        return ("domain={0.domain!r}, address={0.address!r}, "
                "mechanism={0.mechanism!r}, credentials={0.credentials!r}, "
                "user_id={0.user_id!r}, capabilities={0.capabilities!r}".format(self))

    def __repr__(self):
        cls = self.__class__
        return "%s.%s(%s)" % (cls.__module__, cls.__name__, self)

    @staticmethod
    def valid_credentials(cred, mechanism="CURVE"):
        """Raises AuthEntryInvalid if credentials are invalid"""
        AuthEntry.valid_mechanism(mechanism)
        if mechanism == "NULL":
            return
        if cred is None:
            raise AuthEntryInvalid(
                "credentials parameter is required for mechanism {}".format(mechanism))
        if isregex(cred):
            return
        if mechanism == "CURVE" and len(cred) != BASE64_ENCODED_CURVE_KEY_LEN:
            raise AuthEntryInvalid("Invalid CURVE public key {}")

    @staticmethod
    def valid_mechanism(mechanism):
        """Raises AuthEntryInvalid if mechanism is invalid"""
        if mechanism not in ("NULL", "PLAIN", "CURVE"):
            raise AuthEntryInvalid('mechanism must be either "NULL", "PLAIN" or "CURVE"')

    def _check_validity(self):
        """Raises AuthEntryInvalid if entry is invalid"""
        AuthEntry.valid_credentials(self.credentials, self.mechanism)


class AuthFile(object):

    def __init__(self, auth_file=None):
        if auth_file is None:
            auth_file_dir = cc.get_volttron_home()
            auth_file = os.path.join(auth_file_dir, "auth.json")
        self.auth_file = auth_file
        self._check_for_upgrade()

    @property
    def version(self):
        return {"major": 1, "minor": 2}

    def _check_for_upgrade(self):
        allow_list, deny_list, groups, roles, version = self._read()
        if version != self.version:
            if version["major"] <= self.version["major"]:
                self._upgrade(allow_list, deny_list, groups, roles, version)
            else:
                _log.error("This version of VOLTTRON cannot parse {}. "
                           "Please upgrade VOLTTRON or move or delete "
                           "this file.".format(self.auth_file))

    def _read(self):
        auth_data = {}
        try:
            create_file_if_missing(self.auth_file)
            with open(self.auth_file) as fil:
                # Use gevent FileObject to avoid blocking the thread
                before_strip_comments = FileObject(fil, close=False).read()
                if isinstance(before_strip_comments, bytes):
                    before_strip_comments = before_strip_comments.decode("utf-8")
                data = strip_comments(before_strip_comments)
                if data:
                    auth_data = jsonapi.loads(data)
        except Exception:
            _log.exception("error loading %s", self.auth_file)

        allow_list = auth_data.get("allow", [])
        deny_list = auth_data.get("deny", [])
        groups = auth_data.get("groups", {})
        roles = auth_data.get("roles", {})
        version = auth_data.get("version", {"major": 0, "minor": 0})
        return allow_list, deny_list, groups, roles, version

    def read(self):
        """Gets the allowed entries, groups, and roles from the auth
        file.

        :returns: tuple of allow-entries-list, groups-dict, roles-dict
        :rtype: tuple
        """
        allow_list, deny_list, groups, roles, _ = self._read()
        allow_entries, deny_entries = self._get_entries(allow_list, deny_list)
        self._use_groups_and_roles(allow_entries, groups, roles)
        return allow_entries, deny_entries, groups, roles

    def _upgrade(self, allow_list, deny_list, groups, roles, version):
        backup = self.auth_file + "." + str(uuid.uuid4()) + ".bak"
        shutil.copy(self.auth_file, backup)
        _log.info("Created backup of {} at {}".format(self.auth_file, backup))

        def warn_invalid(entry, msg=""):
            _log.warning("Invalid entry {} in auth file {}. {}".format(entry, self.auth_file, msg))

        def upgrade_0_to_1(allow_list):
            new_allow_list = []
            for entry in allow_list:
                try:
                    credentials = entry["credentials"]
                except KeyError:
                    warn_invalid(entry)
                    continue
                if isregex(credentials):
                    msg = "Cannot upgrade entries with regex credentials"
                    warn_invalid(entry, msg)
                    continue
                if credentials == "NULL":
                    mechanism = "NULL"
                    credentials = None
                else:
                    match = re.match(r"^(PLAIN|CURVE):(.*)", credentials)
                    if match is None:
                        msg = "Expected NULL, PLAIN, or CURVE credentials"
                        warn_invalid(entry, msg)
                        continue
                    try:
                        mechanism = match.group(1)
                        credentials = match.group(2)
                    except IndexError:
                        warn_invalid(entry, "Unexpected credential format")
                        continue
                new_allow_list.append({
                    "domain": entry.get("domain"),
                    "address": entry.get("address"),
                    "mechanism": mechanism,
                    "credentials": credentials,
                    "user_id": entry.get("user_id"),
                    "groups": entry.get("groups", []),
                    "roles": entry.get("roles", []),
                    "capabilities": entry.get("capabilities", []),
                    "comments": entry.get("comments"),
                    "enabled": entry.get("enabled", True),
                })
            return new_allow_list

        def upgrade_1_0_to_1_1(allow_list):
            new_allow_list = []
            user_id_set = set()
            for entry in allow_list:
                user_id = entry.get("user_id")
                if user_id:
                    if user_id in user_id_set:
                        new_user_id = str(uuid.uuid4())
                        msg = ("user_id {} is already present in "
                               "authentication entry. Changed to user_id to "
                               "{}").format(user_id, new_user_id)
                        _log.warning(msg)
                        user_id_ = new_user_id
                else:
                    user_id = str(uuid.uuid4())
                user_id_set.add(user_id)
                entry["user_id"] = user_id
                new_allow_list.append(entry)
            return new_allow_list

        def upgrade_1_1_to_1_2(allow_list):
            new_allow_list = []
            for entry in allow_list:
                user_id = entry.get("user_id")
                if user_id in [CONTROL, VOLTTRON_CENTRAL_PLATFORM]:
                    user_id = "/.*/"
                capabilities = entry.get("capabilities")
                entry["capabilities"] = (AuthEntry.build_capabilities_field(capabilities) or {})
                entry["capabilities"]["edit_config_store"] = {"identity": user_id}
                new_allow_list.append(entry)
            return new_allow_list

        if version["major"] == 0:
            allow_list = upgrade_0_to_1(allow_list)
            version["major"] = 1
            version["minor"] = 0
        if version["major"] == 1 and version["minor"] == 0:
            allow_list = upgrade_1_0_to_1_1(allow_list)
            version["minor"] = 1
        if version["major"] == 1 and version["minor"] == 1:
            allow_list = upgrade_1_1_to_1_2(allow_list)

        allow_entries, deny_entries = self._get_entries(allow_list, deny_list)
        self._write(allow_entries, deny_entries, groups, roles)

    def read_allow_entries(self):
        """Gets the allowed entries from the auth file.

        :returns: list of allow-entries
        :rtype: list
        """
        return self.read()[0]

    def read_deny_entries(self):
        """Gets the denied entries from the auth file.

        :returns: list of deny-entries
        :rtype: list
        """
        return self.read()[1]

    def find_by_credentials(self, credentials, is_allow=True):
        """Find all entries that have the given credentials

        :param str credentials: The credentials to search for
        :return: list of entries
        :rtype: list
        """

        if is_allow:
            return [
                entry for entry in self.read_allow_entries()
                if str(entry.credentials) == credentials
            ]
        else:
            return [
                entry for entry in self.read_deny_entries()
                if str(entry.credentials) == credentials
            ]

    def _get_entries(self, allow_list, deny_list):
        allow_entries = []
        for file_entry in allow_list:
            try:
                entry = AuthEntry(**file_entry)
            except TypeError:
                _log.warning("invalid entry %r in auth file %s", file_entry, self.auth_file)
            except AuthEntryInvalid as e:
                _log.warning(
                    "invalid entry %r in auth file %s (%s)",
                    file_entry,
                    self.auth_file,
                    str(e),
                )
            else:
                allow_entries.append(entry)

        deny_entries = []
        for file_entry in deny_list:
            try:
                entry = AuthEntry(**file_entry)
            except TypeError:
                _log.warn("invalid entry %r in auth file %s", file_entry, self.auth_file)
            except AuthEntryInvalid as e:
                _log.warn(
                    "invalid entry %r in auth file %s (%s)",
                    file_entry,
                    self.auth_file,
                    str(e),
                )
            else:
                deny_entries.append(entry)
        return allow_entries, deny_entries

    def _use_groups_and_roles(self, entries, groups, roles):
        """Add capabilities to each entry based on groups and roles"""
        for entry in entries:
            entry_roles = entry.roles
            # Each group is a list of roles
            for group in entry.groups:
                entry_roles += groups.get(group, [])
            capabilities = []
            # Each role is a list of capabilities
            for role in entry_roles:
                capabilities += roles.get(role, [])
            entry.add_capabilities(list(set(capabilities)))

    def _check_if_exists(self, entry, is_allow=True):
        """Raises AuthFileEntryAlreadyExists if entry is already in file"""
        if is_allow:
            for index, prev_entry in enumerate(self.read_allow_entries()):
                if entry.user_id == prev_entry.user_id:
                    raise AuthFileUserIdAlreadyExists(entry.user_id, [index])

                # Compare AuthEntry objects component-wise, rather than
                # using match, because match will evaluate regex.
                if (prev_entry.domain == entry.domain and prev_entry.address == entry.address
                        and prev_entry.mechanism == entry.mechanism
                        and prev_entry.credentials == entry.credentials):
                    raise AuthFileEntryAlreadyExists([index])
        else:
            for index, prev_entry in enumerate(self.read_deny_entries()):
                if entry.user_id == prev_entry.user_id:
                    raise AuthFileUserIdAlreadyExists(entry.user_id, [index])

                # Compare AuthEntry objects component-wise, rather than
                # using match, because match will evaluate regex.
                if (prev_entry.domain == entry.domain and prev_entry.address == entry.address
                        and prev_entry.mechanism == entry.mechanism
                        and prev_entry.credentials == entry.credentials):
                    raise AuthFileEntryAlreadyExists([index])

    def _update_by_indices(self, auth_entry, indices, is_allow=True):
        """Updates all entries at given indices with auth_entry"""
        for index in indices:
            self.update_by_index(auth_entry, index, is_allow)

    def add(self, auth_entry, overwrite=False, no_error=False, is_allow=True):
        """Adds an AuthEntry to the auth file

        :param auth_entry: authentication entry
        :param overwrite: set to true to overwrite matching entries
        :param no_error:
            set to True to not throw an AuthFileEntryAlreadyExists when attempting to add an exiting entry.

        :type auth_entry: AuthEntry
        :type overwrite: bool
        :type no_error: bool

        .. warning:: If overwrite is set to False and if auth_entry matches an
                     existing entry then this method will raise
                     AuthFileEntryAlreadyExists unless no_error is set to true
        """
        try:
            self._check_if_exists(auth_entry, is_allow)
        except AuthFileEntryAlreadyExists as err:
            if overwrite:
                _log.debug("Updating existing auth entry with {} ".format(auth_entry))
                self._update_by_indices(auth_entry, err.indices, is_allow)
            else:
                if not no_error:
                    raise err
        else:
            allow_entries, deny_entries, groups, roles = self.read()
            if is_allow:
                allow_entries.append(auth_entry)
            else:
                deny_entries.append(auth_entry)
            self._write(allow_entries, deny_entries, groups, roles)
            _log.debug("Added auth entry {} ".format(auth_entry))
        gevent.sleep(1)

    def remove_by_credentials(self, credentials, is_allow=True):
        """Removes entry from auth file by credential

        :para credential: entries will this credential will be
            removed
        :type credential: str
        """
        allow_entries, deny_entries, groups, roles = self.read()
        if is_allow:
            entries = allow_entries
        else:
            entries = deny_entries
        entries = [e for e in entries if e.credentials != credentials]
        if is_allow:
            self._write(entries, deny_entries, groups, roles)
        else:
            self._write(allow_entries, entries, groups, roles)

    def remove_by_index(self, index, is_allow=True):
        """Removes entry from auth file by index

        :param index: index of entry to remove
        :type index: int

        .. warning:: Calling with out-of-range index will raise
                     AuthFileIndexError
        """
        self.remove_by_indices([index], is_allow)

    def remove_by_indices(self, indices, is_allow=True):
        """Removes entry from auth file by indices

        :param indices: list of indicies of entries to remove
        :type indices: list

        .. warning:: Calling with out-of-range index will raise
                     AuthFileIndexError
        """
        indices = list(set(indices))
        indices.sort(reverse=True)
        allow_entries, deny_entries, groups, roles = self.read()
        if is_allow:
            entries = allow_entries
        else:
            entries = deny_entries
        for index in indices:
            try:
                del entries[index]
            except IndexError:
                raise AuthFileIndexError(index)
        if is_allow:
            self._write(entries, deny_entries, groups, roles)
        else:
            self._write(allow_entries, entries, groups, roles)

    def _set_groups_or_roles(self, groups_or_roles, is_group=True):
        param_name = "groups" if is_group else "roles"
        if not isinstance(groups_or_roles, dict):
            raise ValueError("{} parameter must be dict".format(param_name))
        for key, value in groups_or_roles.items():
            if not isinstance(value, list):
                raise ValueError("each value of the {} dict must be "
                                 "a list".format(param_name))
        allow_entries, deny_entries, groups, roles = self.read()
        if is_group:
            groups = groups_or_roles
        else:
            roles = groups_or_roles
        self._write(allow_entries, deny_entries, groups, roles)

    def set_groups(self, groups):
        """Define the mapping of group names to role lists

        :param groups: dict where the keys are group names and the
                       values are lists of capability names
        :type groups: dict

        .. warning:: Calling with invalid groups will raise ValueError
        """
        self._set_groups_or_roles(groups, is_group=True)

    def set_roles(self, roles):
        """Define the mapping of role names to capability lists

        :param roles: dict where the keys are role names and the
                      values are lists of group names
        :type groups: dict

        .. warning:: Calling with invalid roles will raise ValueError
        """
        self._set_groups_or_roles(roles, is_group=False)

    def update_by_index(self, auth_entry, index, is_allow=True):
        """Updates entry will given auth entry at given index

        :param auth_entry: new authorization entry
        :param index: index of entry to update
        :type auth_entry: AuthEntry
        :type index: int

        .. warning:: Calling with out-of-range index will raise
                     AuthFileIndexError
        """
        allow_entries, deny_entries, groups, roles = self.read()
        if is_allow:
            entries = allow_entries
        else:
            entries = deny_entries
        try:
            entries[index] = auth_entry
        except IndexError:
            raise AuthFileIndexError(index)
        if is_allow:
            self._write(entries, deny_entries, groups, roles)
        else:
            self._write(allow_entries, entries, groups, roles)

    def _write(self, allow_entries, deny_entries, groups, roles):
        auth = {
            "allow": [vars(x) for x in allow_entries],
            "deny": [vars(x) for x in deny_entries],
            "groups": groups,
            "roles": roles,
            "version": self.version,
        }

        with open(self.auth_file, "w") as fp:
            jsonapi.dump(auth, fp, indent=2)


class AuthFileIndexError(AuthException, IndexError):
    """Exception for invalid indices provided to AuthFile"""

    def __init__(self, indices, message=None):
        if not isinstance(indices, list):
            indices = [indices]
        if message is None:
            message = "Invalid {}: {}".format("indicies" if len(indices) > 1 else "index", indices)
        super(AuthFileIndexError, self).__init__(message)
        self.indices = indices


class AuthFileEntryAlreadyExists(AuthFileIndexError):
    """Exception if adding an entry that already exists"""

    def __init__(self, indicies, message=None):
        if message is None:
            message = ("entry matches domain, address and credentials at "
                       "index {}").format(indicies)
        super(AuthFileEntryAlreadyExists, self).__init__(indicies, message)


class AuthFileUserIdAlreadyExists(AuthFileEntryAlreadyExists):
    """Exception if adding an entry that has a taken user_id"""

    def __init__(self, user_id, indicies, message=None):
        if message is None:
            message = ("user_id {} is already in use at "
                       "index {}").format(user_id, indicies)
        super(AuthFileUserIdAlreadyExists, self).__init__(indicies, message)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import os
import sys
from typing import Optional
from urllib.parse import urlparse
import uuid

import zmq
from zmq import ZMQError, NOBLOCK

from volttron.utils import serialize_frames, deserialize_frames, jsonapi
from volttron.utils.logs import FramesFormatter
from volttron.utils.socket import Address
from volttron.utils.keystore import KeyStore

from .base_router import BaseRouter, UNROUTABLE, ERROR, INCOMING

from volttron.services.routing import ExternalRPCService, PubSubService

from volttron.types.peer import ServicePeerNotifier
from volttron.services.routing import RoutingService
from volttron.server.monitor import Monitor
from ...client.known_identities import CONTROL

_log = logging.getLogger(__name__)


class Router(BaseRouter):
    """Concrete VIP router."""

    def __init__(
        self,
        local_address,
        addresses=(),
        context=None,
        secretkey=None,
        publickey=None,
        default_user_id=None,
        monitor=False,
        tracker=None,
        volttron_central_address=None,
        instance_name=None,
        bind_web_address=None,
        volttron_central_serverkey=None,
        protected_topics={},
        external_address_file="",
        msgdebug=None,
        agent_monitor_frequency=600,
        service_notifier=Optional[ServicePeerNotifier],
    ):

        super(Router, self).__init__(
            context=context,
            default_user_id=default_user_id,
            service_notifier=service_notifier,
        )
        self.local_address = Address(local_address)
        self._addr = addresses
        self.addresses = addresses = [Address(addr) for addr in set(addresses)]
        self._secretkey = secretkey
        self._publickey = publickey
        self.logger = logging.getLogger("vip.router")
        if self.logger.level == logging.NOTSET:
            self.logger.setLevel(logging.DEBUG) # .WARNING)
        self._monitor = monitor
        self._tracker = tracker
        self._volttron_central_address = volttron_central_address
        if self._volttron_central_address:
            parsed = urlparse(self._volttron_central_address)

            assert parsed.scheme in (
                "http",
                "https",
                "tcp",
                "amqp",
            ), "volttron central address must begin with http(s) or tcp found"
            if parsed.scheme == "tcp":
                assert (volttron_central_serverkey
                        ), "volttron central serverkey must be set if address is tcp."
        self._volttron_central_serverkey = volttron_central_serverkey
        self._instance_name = instance_name
        self._bind_web_address = bind_web_address
        self._protected_topics = protected_topics
        self._external_address_file = external_address_file
        self._pubsub = None
        self.ext_rpc = None
        self._msgdebug = msgdebug
        self._message_debugger_socket = None
        self._instance_name = instance_name
        self._agent_monitor_frequency = agent_monitor_frequency

    def setup(self):
        sock = self.socket
        identity = str(uuid.uuid4())
        sock.identity = identity.encode("utf-8")
        _log.debug("ROUTER SOCK identity: {}".format(sock.identity))
        if self._monitor:
            Monitor(sock.get_monitor_socket()).start()
        sock.bind("inproc://vip")
        _log.debug("In-process VIP router bound to inproc://vip")
        sock.zap_domain = b"vip"
        addr = self.local_address
        if not addr.identity:
            addr.identity = identity
        if not addr.domain:
            addr.domain = "vip"

        addr.server = "CURVE"
        addr.secretkey = self._secretkey

        addr.bind(sock)
        _log.debug("Local VIP router bound to %s" % addr)
        for address in self.addresses:
            if not address.identity:
                address.identity = identity
            if (address.secretkey is None and address.server not in ["NULL", "PLAIN"]
                    and self._secretkey):
                address.server = "CURVE"
                address.secretkey = self._secretkey
            if not address.domain:
                address.domain = "vip"
            address.bind(sock)
            _log.debug("Additional VIP router bound to %s" % address)
        self._ext_routing = None

        self._ext_routing = RoutingService(
            self.socket,
            self.context,
            self._socket_class,
            self._poller,
            self._addr,
            self._instance_name,
        )

        self.pubsub = PubSubService(self.socket, self._protected_topics, self._ext_routing)
        self.ext_rpc = ExternalRPCService(self.socket, self._ext_routing)
        self._poller.register(sock, zmq.POLLIN)
        _log.debug("ZMQ version: {}".format(zmq.zmq_version()))

    def issue(self, topic, frames, extra=None):
        log = self.logger.debug
        formatter = FramesFormatter(frames)
        if topic == ERROR:
            errnum, errmsg = extra
            log("%s (%s): %s", errmsg, errnum, formatter)
        elif topic == UNROUTABLE:
            log("unroutable: %s: %s", extra, formatter)
        else:
            direction = "incoming" if topic == INCOMING else "outgoing"
            if direction == "outgoing":
                log(f"{direction}: {deserialize_frames(frames)}")
            else:
                log(f"{direction}: {frames}")
        if self._tracker:
            self._tracker.hit(topic, frames, extra)
        if self._msgdebug:
            if not self._message_debugger_socket:
                # Initialize a ZMQ IPC socket on which to publish all messages to MessageDebuggerAgent.
                socket_path = os.path.expandvars("$VOLTTRON_HOME/run/messagedebug")
                socket_path = os.path.expanduser(socket_path)
                socket_path = ("ipc://{}".format("@" if sys.platform.startswith("linux") else "") +
                               socket_path)
                self._message_debugger_socket = zmq.Context().socket(zmq.PUB)
                self._message_debugger_socket.connect(socket_path)
            # Publish the routed message, including the "topic" (status/direction), for use by MessageDebuggerAgent.
            frame_bytes = [topic]
            frame_bytes.extend(
                frames)    # [frame if type(frame) is bytes else frame.bytes for frame in frames])
            frame_bytes = serialize_frames(frames)
            # TODO we need to fix the msgdebugger socket if we need it to be connected
            # frame_bytes = [f.bytes for f in frame_bytes]
            # self._message_debugger_socket.send_pyobj(frame_bytes)

    # This is currently not being used e.g once fixed we won't use it.
    # def extract_bytes(self, frame_bytes):
    #    result = []
    #    for f in frame_bytes:
    #        if isinstance(f, list):
    #            result.extend(self.extract_bytes(f))
    #        else:
    #            result.append(f.bytes)
    #    return result

    def handle_subsystem(self, frames, user_id):
        _log.debug(f"Handling subsystem with frames: {frames} user_id: {user_id}")

        subsystem = frames[5]
        if subsystem == "quit":
            sender = frames[0]
            # was if sender == 'control' and user_id == self.default_user_id:
            # now we serialize frames and if user_id is always the sender and not
            # recipents.get('User-Id') or default user name
            if sender == CONTROL:
                if self._ext_routing:
                    self._ext_routing.close_external_connections()
                self.stop()
                raise KeyboardInterrupt()
            else:
                _log.error(f"Sender {sender} not authorized to shutdown platform")
        elif subsystem == "agentstop":
            try:
                drop = frames[6]
                self._drop_peer(drop)
                self._drop_pubsub_peers(drop)
                if self._service_notifier:
                    self._service_notifier.peer_dropped(drop)

                _log.debug("ROUTER received agent stop message. dropping peer: {}".format(drop))
            except IndexError:
                _log.error(
                    f"agentstop called but unable to determine agent from frames sent {frames}")
            return False
        elif subsystem == "query":
            try:
                name = frames[6]
            except IndexError:
                value = None
            else:
                if name == "addresses":
                    if self.addresses:
                        value = [addr.base for addr in self.addresses]
                    else:
                        value = [self.local_address.base]
                elif name == "local_address":
                    value = self.local_address.base
                # Allow the agents to know the serverkey.
                elif name == "serverkey":
                    keystore = KeyStore()
                    value = keystore.public
                elif name == "volttron-central-address":
                    value = self._volttron_central_address
                elif name == "volttron-central-serverkey":
                    value = self._volttron_central_serverkey
                elif name == "instance-name":
                    value = self._instance_name
                elif name == "bind-web-address":
                    value = self._bind_web_address
                elif name == "platform-version":
                    raise NotImplementedError()
                    # value = __version__
                elif name == "message-bus":
                    value = os.environ.get("MESSAGEBUS", "zmq")
                elif name == "agent-monitor-frequency":
                    value = self._agent_monitor_frequency
                else:
                    value = None
            frames[6:] = ["", value]
            frames[3] = ""

            return frames
        elif subsystem == "pubsub":
            result = self.pubsub.handle_subsystem(frames, user_id)
            return result
        elif subsystem == "routing_table":
            result = self._ext_routing.handle_subsystem(frames)
            return result
        elif subsystem == "external_rpc":
            result = self.ext_rpc.handle_subsystem(frames)
            return result

    def _drop_pubsub_peers(self, peer):
        self.pubsub.peer_drop(peer)

    def _add_pubsub_peers(self, peer):
        self.pubsub.peer_add(peer)

    def poll_sockets(self):
        """
        Poll for incoming messages through router socket or other external socket connections
        """
        try:
            sockets = dict(self._poller.poll())
        except ZMQError as ex:
            _log.error("ZMQ Error while polling: {}".format(ex))

        for sock in sockets:
            if sock == self.socket:
                if sockets[sock] == zmq.POLLIN:
                    frames = sock.recv_multipart(copy=False)
                    self.route(deserialize_frames(frames))
            elif sock in self._ext_routing._vip_sockets:
                if sockets[sock] == zmq.POLLIN:
                    # _log.debug("From Ext Socket: ")
                    self.ext_route(sock)
            elif sock in self._ext_routing._monitor_sockets:
                self._ext_routing.handle_monitor_event(sock)
            else:
                # _log.debug("External ")
                frames = sock.recv_multipart(copy=False)

    def ext_route(self, socket):
        """
        Handler function for message received through external socket connection
        :param socket: socket affected files: {}
        :return:
        """
        # Expecting incoming frames to follow this VIP format:
        #   [SENDER, PROTO, USER_ID, MSG_ID, SUBSYS, ...]
        frames = socket.recv_multipart(copy=False)
        self.route(deserialize_frames(frames))
        # for f in frames:
        #     _log.debug("PUBSUBSERVICE Frames: {}".format(bytes(f)))
        if len(frames) < 6:
            return

        sender, proto, user_id, msg_id, subsystem = frames[:5]
        if proto != "VIP1":
            return

        # Handle 'EXT_RPC' subsystem messages
        name = subsystem
        if name == "external_rpc":
            # Reframe the frames
            sender, proto, usr_id, msg_id, subsystem, msg = frames[:6]
            msg_data = jsonapi.loads(msg)
            peer = msg_data["to_peer"]
            # Send to destionation agent/peer
            # Form new frame for local
            frames[:9] = [
                peer,
                sender,
                proto,
                usr_id,
                msg_id,
                "external_rpc",
                msg,
            ]
            try:
                self.socket.send_multipart(frames, flags=NOBLOCK, copy=False)
            except ZMQError as ex:
                _log.debug("ZMQ error: {}".format(ex))
                pass
        # Handle 'pubsub' subsystem messages
        elif name == "pubsub":
            if frames[1] == "VIP1":
                recipient = ""
                frames[:1] = ["", ""]
                # for f in frames:
                #     _log.debug("frames: {}".format(bytes(f)))
            result = self.pubsub.handle_subsystem(frames, user_id)
            return result
        # Handle 'routing_table' subsystem messages
        elif name == "routing_table":
            # for f in frames:
            #     _log.debug("frames: {}".format(bytes(f)))
            if frames[1] == "VIP1":
                frames[:1] = ["", ""]
            result = self._ext_routing.handle_subsystem(frames)
            return result



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import random
from weakref import WeakValueDictionary

from gevent.event import AsyncResult
import sys
from datetime import datetime

__all__ = ["counter", "ResultsDictionary"]


class AsyncResult(AsyncResult):
    __slots__ = AsyncResult.__slots__ + ("ident", )


def counter(start=None, minimum=0, maximum=sys.maxsize - 1):
    # count = random.randint(minimum, maximum) if start is None else start
    count = int(datetime.now().timestamp()) if start is None else start
    while True:
        yield count
        count += 1
        if count >= maximum:
            count = int(datetime.now().timestamp())


class ResultsDictionary(WeakValueDictionary):

    def __init__(self):
        WeakValueDictionary.__init__(self)
        self._counter = counter()

    def pop(self, key, *args):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.pop(self, key, *args)

    def __contains__(self, key):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.__contains__(self, key)

    def __getitem__(self, key):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.__getitem__(self, key)

    def get(self, key, default=None):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.get(self, key, default=default)

    def __next__(self):
        result = AsyncResult()
        result.ident = ident = "%f.%f" % (next(self._counter), hash(result))
        self[ident] = result
        return result



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

"""VIP - VOLTTRON™ Interconnect Protocol implementation

See https://volttron.readthedocs.io/en/develop/core_services/messagebus/VIP/VIP-Overview.html
for protocol specification.

This module is useful for using VIP outside of gevent. Please understand
that ZeroMQ sockets are not thread-safe and care must be used when using
across threads (or avoided all together). There is no locking around the
state as there is with the gevent version in the green sub-module.
"""

from threading import local as _local

import zmq as _zmq

from volttron.utils.socket import _Socket


class Socket(_Socket, _zmq.Socket):
    _context_class = _zmq.Context
    _local_class = _local


class BaseConnection(object):
    """
    Base connection class for message bus connection.
    """

    def __init__(self, url, identity, instance_name):
        self._url = url
        self._identity = identity
        self._instance_name = instance_name
        self._vip_handler = None



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import functools
import logging
import random
import string
import weakref

import gevent
from zmq import green as zmq
from zmq import ZMQError

from volttron.utils.frame_serialization import serialize_frames
from .base import SubsystemBase

_log = logging.getLogger(__name__)
_log.setLevel(logging.WARN)

__all__ = ['Channel']


class Channel(SubsystemBase):
    """
    A low level interface for handling a protocol that is not based upon vip.  This system
    will setup a channel between itself and a peer.  The socket will be handed back to the
    creator and can be used to send messages via the zmq.Socket interface.
    """
    ADDRESS = 'inproc://subsystem/channel'

    def __init__(self, core):
        self.context = zmq.Context()
        self.socket = None
        self.greenlet = None
        self._channels = {}
        core.register('channel', self._handle_subsystem)

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self.socket = self.context.socket(zmq.ROUTER)

        core.onsetup.connect(setup, self)

        def start(sender, **kwargs):
            # pylint: disable=unused-argument
            self.greenlet = gevent.getcurrent()
            vip_sock = core.socket
            chan_sock = self.socket
            chan_sock.bind(self.ADDRESS)
            while True:
                message = chan_sock.recv_multipart(copy=False)
                if not message:
                    continue
                ident = bytes(message[0])
                try:
                    peer, name = self._channels[ident]
                except KeyError:
                    # XXX: Handle channel not found
                    continue
                message[0] = name
                # _log.debug(f"Sending vip through channel {message}")
                vip_sock.send_vip(peer, 'channel', message, copy=False)

        core.onstart.connect(start, self)

        def stop(sender, **kwargs):
            # pylint: disable=unused-argument
            if self.greenlet is not None:
                self.greenlet.kill(block=False)
            try:
                self.socket.unbind(self.ADDRESS)
            except ZMQError:
                pass

        core.onstop.connect(stop, self)

    def _handle_subsystem(self, message):
        """
        Handle the channel subsystem sending multipart data through the socket.
        """
        _log.debug(f"Message received: {message}")
        frames = message.args
        try:
            name = frames[0]
        except IndexError:
            return
        channel = (message.peer, name)
        try:
            ident = self._channels[channel]
        except KeyError:
            # XXX: Handle channel not found
            _log.error(f"Channel {channel} not found!")
            return
        frames[0] = ident

        try:
            memoryview(frames)
            _log.debug("Serializing frames not necessary")
        except TypeError:
            _log.debug("Serializing frames necessary")
            frames = serialize_frames(frames)

        _log.debug(f"frames are before send_multipart: {frames}")

        self.socket.send_multipart(frames, copy=False)

    def create(self, peer, name=None) -> zmq.Socket:
        """
        Create a zmq socket between the caller agent and a peer.  This channel allows a protocol
        to be set up to pass binary data back and forth between the two agent instances without
        having to deal with the vip protocol itself.
        """

        if name is None:
            while True:
                name = ''.join(random.choice(string.printable[:-5]) for i in range(30))
                channel = (peer, name)
                if channel not in self._channels:
                    break
        else:
            channel = (peer, name)
            if channel in self._channels:
                raise ValueError('channel %r is unavailable' % (name, ))
        _log.debug(f"Connecting to channel {channel}")
        sock = self.context.socket(zmq.DEALER)
        sock.hwm = 1
        sock.identity = ident = f'{hash(channel)}s.{hash(sock)}s'.encode('utf-8')
        sockref = weakref.ref(sock, self._destroy)
        object.__setattr__(sock, 'peer', peer)
        object.__setattr__(sock, 'name', name)
        self._channels[channel] = ident
        self._channels[ident] = channel
        self._channels[sockref] = (ident, peer, name)
        close_socket = sock.close

        @functools.wraps(close_socket)
        def close(linger=None):
            self._destroy(sockref)
            sock.close = close_socket
            return close_socket(linger=linger)

        sock.close = close
        sock.connect(self.ADDRESS)
        return sock

    __call__ = create

    def _destroy(self, sockref):
        try:
            ident, peer, name = self._channels.pop(sockref)
        except KeyError:
            return
        self._channels.pop(ident, None)
        self._channels.pop((peer, name), None)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import base64
from contextlib import closing
import csv
import hashlib
import logging
import os
import posixpath
import re
import shutil
import sys
import tempfile
import zipfile

from volttron.utils import jsonapi

# from wheel.install import WheelFile
# from wheel.util import native, open_for_csv

__all__ = (
    "BasePackageVerifier",
    "VolttronPackageWheelFileNoSign",
    "ZipPackageVerifier",
    "UnpackedPackage",
)

_log = logging.getLogger(__name__)


# TODO: Make this this base class and have auth extend it
class VolttronPackageWheelFileNoSign:
    AGENT_DATA_ZIP = "agent_data.zip"

    def __init__(self, filename, **kwargs):

        super(VolttronPackageWheelFileNoSign, self).__init__(filename, **kwargs)

    def contains(self, path):
        """Does the wheel contain the specified path?"""

        for x in self.zipfile.filelist:
            if x.filename == path:
                return True
        return False

    def add_files(self, files_to_add=None, basedir="."):

        if files_to_add == None or len(files_to_add) == 0:
            return

        records = ZipPackageVerifier(self.filename).get_records()

        if len(records) < 1:
            raise ValueError("Invalid wheel file no records found")

        last_record_name = records[0]
        #         new_record_name = "RECORD.{}".format(len(records))
        #
        tmp_dir = tempfile.mkdtemp()
        try:
            record_path = "/".join((self.distinfo_name, last_record_name))
            tmp_new_record_file = "/".join((tmp_dir, self.distinfo_name, last_record_name))
            self.zipfile.extract("/".join((self.distinfo_name, last_record_name)), path=tmp_dir)

            self.remove_files("/".join((self.distinfo_name, "config")))

            with closing(open_for_csv(tmp_new_record_file, "a+")) as record_file:
                writer = csv.writer(record_file)

                if files_to_add:
                    if "config_file" in files_to_add:
                        try:
                            data = open(files_to_add["config_file"]).read()
                        except OSError as e:
                            _log.error("couldn't access {}" % files_to_add["config_file"])
                            raise

                        self.zipfile.writestr("%s/%s" % (self.distinfo_name, "config"), data)

                        (hash_data, size, digest) = self._record_digest(data)
                        record_path = "/".join((self.distinfo_name, "config"))
                        writer.writerow((record_path, hash_data, size))

                    if "identity_file" in files_to_add:
                        try:
                            data = open(files_to_add["identity_file"]).read()
                        except OSError as e:
                            _log.error("couldn't access {}" % files_to_add["identity_file"])
                            raise

                        self.zipfile.writestr("%s/%s" % (self.distinfo_name, "IDENTITY_TEMPLATE"),
                                              data)

                        (hash_data, size, digest) = self._record_digest(data)
                        record_path = "/".join((self.distinfo_name, "IDENTITY_TEMPLATE"))
                        writer.writerow((record_path, hash_data, size))

                    if ("contract" in files_to_add and files_to_add["contract"] is not None):
                        try:
                            data = open(files_to_add["contract"]).read()
                        except OSError as e:
                            _log.error("couldn't access {}" % files_to_add["contract"])
                            raise

                        if files_to_add["contract"] != "execreqs.json":
                            msg = "WARNING: renaming passed contract file: {}".format(
                                files_to_add["contract"])
                            msg += " to execreqs.json"
                            sys.stderr.write(msg)
                            _log.warning(msg)

                        self.zipfile.writestr("%s/%s" % (self.distinfo_name, "execreqs.json"),
                                              data)
                        (hash_data, size, digest) = self._record_digest(data)
                        record_path = "/".join((self.distinfo_name, "execreqs.json"))
                        writer.writerow((record_path, hash_data, size))

                    self.__setupzipfile__()

            self.pop_records_file()

            new_record_content = open(tmp_new_record_file, "r").read()
            self.zipfile.writestr(self.distinfo_name + "/" + last_record_name, new_record_content)

            self.zipfile.close()
            self.__setupzipfile__()
        finally:
            shutil.rmtree(tmp_dir, True)

    def pop_records_file(self):
        """Pop off the last records file that was added"""
        records = ZipPackageVerifier(self.filename).get_records()
        topop = (os.path.join(self.distinfo_name, records[0]), )
        self.remove_files(topop)

    def pop_record_and_files(self):
        """Pop off the last record file and files listed in it.

        Only removes files that are not listed in remaining records.
        """
        records = ZipPackageVerifier(self.filename).get_records()
        record = records.pop(0)
        zf = self.zipfile
        keep = set(row[0] for name in records
                   for row in csv.reader(zf.open(posixpath.join(self.distinfo_name, name))))
        drop = set(row[0]
                   for row in csv.reader(zf.open(posixpath.join(self.distinfo_name, record))))
        # These two should already be listed, but add them just in case
        drop.add(posixpath.join(self.distinfo_name, record))
        self.remove_files(drop - keep)

    def remove_files(self, files):
        """Relative to files in the package, ie: ./dist-info/config."""
        if isinstance(files, str):
            files = [files]
        tmpdir = tempfile.mkdtemp()
        try:
            newzip = zipfile.ZipFile(os.path.join(tmpdir, "tmp.zip"), "w")
            with newzip:
                for f in self.zipfile.infolist():
                    if f.filename not in files:
                        buf = self.zipfile.read(f.filename)
                        newzip.writestr(f.filename, buf)
            self.zipfile.close()
            self.fp = None
            os.remove(self.filename)
            shutil.move(newzip.filename, self.filename)
            self.__setupzipfile__()
        finally:
            shutil.rmtree(tmpdir, True)

    def unpack(self, dest="."):
        namever = self.parsed_filename.group("namever")
        destination = os.path.join(dest, namever)
        sys.stderr.write("Unpacking to: %s\n" % (destination))
        self.zipfile.extractall(destination)
        self.zipfile.close()

    def _record_digest(self, data):
        """Returns a three tuple of hash, size and digest."""

        from wheel.util import urlsafe_b64encode

        digest = hashlib.sha256(data.encode("utf-8")).digest()
        hash_text = "sha256=" + native(urlsafe_b64encode(digest))
        size = len(data)
        return (hash_text, size, digest)

    def __setupzipfile__(self):
        self.zipfile.close()
        self.fp = None

        mode = "r"
        if self.append:
            mode = "a"

        self.zipfile = zipfile.ZipFile(
            self.filename,
            mode=mode,
        )


_record_re = re.compile(r"^RECORD(?:\.\d+)?$")
_all_record_re = re.compile(r"^(?:.*/)?RECORD(?:\.\d+)?(?:\.p7s)?$")


#
# Signature verification in the class below has the limitation that only
# a single certificate may be used for verification. Ideally, the
# certificate should be extracted from the signature file and verified
# against a certificate authority (CA) in the CA store. See
# http://code.activestate.com/recipes/285211/ for an alternate solution
# using M2Crypto.
#
class BasePackageVerifier(object):
    """Base class for implementing wheel package verification.

    Verifies wheel packages as defined in PEP-427. May be inherited with
    minimal modifications to support different storage mechanisms, such
    as a filesystem, Zip file, or tarball. All paths are expected to be
    POSIX-style with forward-slashes. Subclasses should implement
    listdir and open and may override __init__, if needed.

    As an extension of the original specification, multiple levels of
    RECORD files and signatures are supported by appending incrementing
    integers to the RECORD files. Verification happens in reverse order
    and later RECORD files should contain hashes of the previous RECORD
    and associated signature file(s).
    """

    def __init__(self, dist_info, **kwargs):
        """Initialize the instance with the dist-info directory name.

        dist_info should contain the name of a single directory, not a
        multi-component path.
        """
        self.dist_info = dist_info

    def listdir(self, path):
        """Return a possibly empty list of files from a directory.

        This could return the contents of a directory in an archive or
        whatever makes sense for the storage mechanism. Paths will
        typically be relative to the package, however, for installed
        packages, absolute paths are possible.
        """
        raise NotImplementedError()

    def open(self, path, mode="r"):
        """Return a file-like object for the given file.

        mode is interpreted the same as for the built-in open and will
        be either 'r' or 'rb'. Only the __iter__(), read(), and close()
        methods are used.
        """
        raise NotImplementedError()

    def iter_hashes(self, name="RECORD"):
        """Iterate over the files and hashes of a RECORD file.

        The RECORD file with the given name will be iterated over
        yielding a three tuple with each iteration: filename (relative
        to the package), computed hash (just calculated), and expected
        hash (from RECORD file).
        """
        hashless = [posixpath.join(self.dist_info, name + ext) for ext in ["", ".jws", ".p7s"]]
        path = posixpath.join(self.dist_info, name)
        with closing(self.open(path)) as record_file:
            for row in csv.reader(record_file):
                filename, hashspec = row[:2]
                if not hashspec:
                    if filename not in hashless:
                        yield filename, None, None
                    continue
                algo, expected_hash = hashspec.split("=", 1)
                hash = hashlib.new(algo)
                with closing(self.open(filename, "rb")) as file:
                    while True:
                        data = file.read(4096)
                        if not data:
                            break
                        hash.update(data)
                hash = base64.urlsafe_b64encode(hash.digest()).rstrip("=")
                yield filename, hash, expected_hash

    def get_records(self):
        """Return a reverse sorted list of RECORD names from the package.

        Returns all RECORD files in the dist_info directory.
        """
        records = [name for name in self.listdir(self.dist_info) if _record_re.match(name)]
        records.sort(key=lambda x: int((x.split(".", 1) + [-1])[1]), reverse=True)
        if not records:
            raise ValueError("missing RECORD file(s) in .dist-info directory")
        return records


class ZipPackageVerifier(BasePackageVerifier):
    """Verify files of a Zip file."""

    def __init__(self, zip_path, mode="r", **kwargs):
        self._zipfile = zipfile.ZipFile(zip_path)
        self._namelist = self._zipfile.namelist()

        names = [
            name for name in self._namelist
            if name.endswith(".dist-info/RECORD") and name.count("/") == 1
        ]
        if len(names) != 1:
            raise ValueError("unable to determine dist-info directory")
        dist_info = names[0].split("/", 1)[0]
        super(ZipPackageVerifier, self).__init__(dist_info, **kwargs)

    def listdir(self, path):
        if path[-1:] != "/":
            path += "/"
        n = len(path)
        return [name[n:].split("/", 1)[0] for name in self._namelist if name.startswith(path)]

    def open(self, path, mode="r"):
        return self._zipfile.open(path, "r")


class UnpackedPackage(object):
    """Represents an package unpacked into a directory.

    Allows one access to the package metadata and methods to repack.
    """

    def __init__(self, base_directory):
        self.directory = base_directory
        self.distinfo = self._get_dist_info()

    def _get_dist_info(self):
        """Find the package.dist-info directory."""
        basename = os.path.basename(self.directory)
        path = os.path.join(self.directory, basename + ".dist-info")
        if os.path.exists(path):
            return path
        for name in os.listdir(self.directory):
            if not name.endswith(".dist-info"):
                continue
            return os.path.join(self.directory, name)
        raise ValueError("directory does not contain a valid "
                         "agent package: {}".format(self.directory))

    @property
    def metadata(self):
        """Parse package.dist-info/metadata.json and return a dictionary."""
        try:
            return self._metadata
        except AttributeError:
            with open(os.path.join(self.distinfo, "metadata.json")) as file:
                self._metadata = jsonapi.load(file)
        return self._metadata

    @property
    def wheelmeta(self):
        """Parse package.dist-info/WHEEL and return a dictionary."""
        try:
            return self._wheelmeta
        except AttributeError:
            with open(os.path.join(self.distinfo, "WHEEL")) as file:
                self._wheelmeta = {
                    key.strip().lower(): value.strip()
                    for key, value in (parts for line in file if line
                                       for parts in [line.split(":", 1)] if len(parts) == 2)
                }
        return self._wheelmeta

    @property
    def package_name(self):
        metadata = self.metadata
        name = metadata["name"]
        version = metadata["version"]
        return "-".join([name, version])

    @property
    def version(self):
        metadata = self.metadata
        return metadata["version"]

    @property
    def wheel_name(self):
        metadata = self.metadata
        name = metadata["name"]
        version = metadata["version"]
        tag = self.wheelmeta["tag"]
        return "-".join([name, version, tag]) + ".whl"

    def repack(self, dest=None, exclude=None):
        """Recreate the package from the RECORD files.

        Put the package in the directory given by dest or in the current
        directory if dest is None. If exclude is given, do not add files
        for RECORD files in exclude. Returns the path to the new package.
        """
        # Get a list of the record files and sort them ascending
        records = [name for name in os.listdir(self.distinfo) if _record_re.match(name)]
        records.sort()
        wheelname = self.wheel_name
        if dest is not None:
            dest = os.path.expanduser(os.path.expandvars(dest))
            wheelname = os.path.join(dest, wheelname)
        # Recreate the package
        with zipfile.ZipFile(wheelname, "w") as wheelfile:
            for record in records:
                # Exclude and record files in exclude
                if exclude and record in exclude:
                    continue
                with open(os.path.join(self.distinfo, record)) as file:
                    csvfile = csv.reader(file)
                    for row in csvfile:
                        name = row[0]
                        # Skip already added RECORD files or signatures
                        if _all_record_re.match(name) and name in wheelfile.namelist():
                            continue
                        wheelfile.write(os.path.join(self.directory, name), name)
        return wheelfile.filename



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

"""Math functions for VOLTTRON platform™ agents.

Not meant to replace numpy in all cases. A basic set common math
routines to remove the need for numpy in simple cases.

This module should NEVER import numpy as that would defeat the
purpose.
"""

from math import sqrt


def mean(data):
    """Return the sample arithmetic mean of data."""
    n = len(data)
    if n < 1:
        raise ValueError('mean requires at least one data point')
    return sum(data) / n    # in Python 2 use sum(data)/float(n)


def _ss(data):
    """Return sum of square deviations of sequence data."""
    c = mean(data)
    ss = sum((x - c)**2 for x in data)
    return ss


def pstdev(data):
    """Calculate the population standard deviation."""
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss / n    # the population variance
    return sqrt(pvar)


def stdev(data):
    """Calculate the sample standard deviation."""
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss / (n - 1)    # sample variance
    return sqrt(pvar)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import weakref

from .base import SubsystemBase
from ..results import ResultsDictionary
from zmq import ZMQError
from zmq.green import ENOTSOCK

__all__ = ["Hello"]

_log = logging.getLogger(__name__)


class Hello(SubsystemBase):
    """The hello subsystem allows an agent to determine its identity.

    The identity is possibly a dynamically generated uuid from which the
    executing agent does not know.  This subsystem allows the agent to be
    able to determine it's identity from a peer.  By default that peer is
    the connected router, however this could be another agent.
    """

    def __init__(self, core):
        self.core = weakref.ref(core)
        self._results = ResultsDictionary()
        core.register("hello", self._handle_hello, self._handle_error)

    def hello(self, peer=""):
        """Receives a welcome message from the peer (default to '' router)

         The welcome message will respond with a 3 element list:

         - The vip version (default 1.0)
         - The peer who responded (should be the same as `peer` argument
           to this function.
         - The id of the requester (i.e. this object).  This will be the
           identity when the agent connects to the router or the specified
           identity when the `Agent` is constructed.

        :param peer: The peer to receive the response from.
        :return: [version, peer, identity]
        """
        _log.info("{0} Requesting hello from peer ({1})".format(self.core().identity, peer))
        result = next(self._results)
        connection = self.core().connection
        if not connection:
            _log.error("Connection object not yet created".format(self.core().identity))
        else:
            try:
                connection.send_vip(peer, "hello", args=["hello"], msg_id=result.ident)
            except ZMQError as exc:
                if exc.errno == ENOTSOCK:
                    _log.error("Socket send on non socket {}".format(self.core().identity))

        return result

    __call__ = hello

    def _handle_hello(self, message):
        _log.info("Handling hello message {}".format(message))
        try:
            # zmq
            op = message.args[0]
        except IndexError:
            _log.error("missing hello subsystem operation")
            return
        if op == "hello":
            message.user = ""
            message.args = ["welcome", "1.0", self.core.identity, message.peer]
            self.core().connection.send_vip_object(message, copy=False)
        elif op == "welcome":
            try:
                result = self._results.pop(message.id)
            except KeyError:
                return
            result.set([arg for arg in message.args[1:]])
        else:
            _log.error("unknown hello subsystem operation")

    def _handle_error(self, sender, message, error, **kwargs):
        try:
            result = self._results.pop(message.id)
        except KeyError:
            return
        result.set_exception(error)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from base64 import b64encode, b64decode
import inspect
import logging
import random
import re
import weakref
import gevent

from zmq import green as zmq
from zmq import SNDMORE
from volttron.utils import jsonapi
from .base import SubsystemBase
from ..decorators import annotate, annotations, dualmethod, spawn
from ..errors import Unreachable
from volttron.utils import jsonrpc
from ..results import ResultsDictionary
from gevent.queue import Queue
from collections import defaultdict

__all__ = ["PubSub"]

min_compatible_version = "3.0"
max_compatible_version = ""

# utils.setup_logging()
_log = logging.getLogger(__name__)


def encode_peer(peer):
    if peer.startswith("\x00"):
        return peer[:1] + b64encode(peer[1:])
    return peer


def decode_peer(peer):
    if peer.startswith("\x00"):
        return peer[:1] + b64decode(peer[1:])
    return peer


class PubSub(SubsystemBase):
    """
    Pubsub subsystem concrete class implementation for ZMQ message bus.
    """

    def __init__(self, core, rpc_subsys, peerlist_subsys, owner):
        self.core = weakref.ref(core)
        self.rpc = weakref.ref(rpc_subsys)
        self.peerlist = weakref.ref(peerlist_subsys)
        self._owner = owner

        def platform_subscriptions():
            return defaultdict(subscriptions)

        def subscriptions():
            return defaultdict(set)

        self._my_subscriptions = defaultdict(platform_subscriptions)
        self.protected_topics = ProtectedPubSubTopics()
        core.register("pubsub", self._handle_subsystem, self._handle_error)
        self.vip_socket = None
        self._results = ResultsDictionary()
        self._event_queue = Queue()
        self._retry_period = 300.0
        self._processgreenlet = None

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self._processgreenlet = gevent.spawn(self._process_loop)
            core.onconnected.connect(self._connected)
            self.vip_socket = self.core().socket

            def subscribe(member):    # pylint: disable=redefined-outer-name
                for peer, bus, prefix, all_platforms, queue in annotations(
                        member, set, "pubsub.subscriptions"):
                    # XXX: needs updated in light of onconnected signal
                    self._add_subscription(prefix, member, bus, all_platforms)
                    # _log.debug("SYNC ZMQ: all_platforms {}".format(self._my_subscriptions['internal'][bus][prefix]))

            inspect.getmembers(owner, subscribe)

        core.onsetup.connect(setup, self)

    def _connected(self, sender, **kwargs):
        """
        Synchronize local subscriptions with PubSubService upon receiving connected signal.
        param sender: identity of sender
        type sender: str
        param kwargs: optional arguments
        type kwargs: pointer to arguments
        """
        self.synchronize()

    def _process_callback(self, sender, bus, topic, headers, message):
        """Handle incoming subscription pushes from PubSubService. It iterates over all subscriptions to find the
        subscription matching the topic and bus. It then calls the corresponding callback on finding a match.
        param sender: identity of the publisher
        type sender: str
        param bus: bus
        type bus: str
        param topic: publishing topic
        type topic: str
        param headers: header information for the incoming message
        type headers: dict
        param message: actual message
        type message: dict
        """
        peer = "pubsub"

        handled = 0
        for platform in self._my_subscriptions:
            # _log.debug("SYNC: process callback subscriptions: {}".format(self._my_subscriptions[platform][bus]))
            buses = self._my_subscriptions[platform]
            if bus in buses:
                subscriptions = buses[bus]
                for prefix, callbacks in subscriptions.items():
                    if topic.startswith(prefix):
                        handled += 1
                        for callback in callbacks:
                            callback(peer, sender, bus, topic, headers, message)
        if not handled:
            # No callbacks for topic; synchronize with sender
            self.synchronize()

    def _viperror(self, sender, error, **kwargs):
        if isinstance(error, Unreachable):
            self._peer_drop(self, error.peer)

    def _peer_add(self, sender, peer, **kwargs):
        # Delay sync by some random amount to prevent reply storm.
        delay = random.random()
        self.core().spawn_later(delay, self.synchronize, peer)

    def _peer_drop(self, sender, peer, **kwargs):
        self._sync(peer, {})

    def _sync(self, peer, items):
        items = {(bus, prefix) for bus, topics in items.items() for prefix in topics}
        remove = []
        for bus, subscriptions in self._my_subscriptions.items():
            for prefix, subscribers in subscriptions.items():
                item = bus, prefix
                try:
                    items.remove(item)
                except KeyError:
                    subscribers.discard(peer)
                    if not subscribers:
                        remove.append(item)
                else:
                    subscribers.add(peer)
        for bus, prefix in remove:
            subscriptions = self._my_subscriptions[bus]
            assert not subscriptions.pop(prefix)
        for bus, prefix in items:
            self._add_peer_subscription(peer, bus, prefix)

    def _add_peer_subscription(self, peer, bus, prefix):
        try:
            subscriptions = self._my_subscriptions[bus]
        except KeyError:
            self._my_subscriptions[bus] = subscriptions = dict()
        try:
            subscribers = subscriptions[prefix]
        except KeyError:
            subscriptions[prefix] = subscribers = set()
        subscribers.add(peer)

    def _distribute(self, peer, topic, headers, message=None, bus=""):
        self._check_if_protected_topic(topic)
        try:
            subscriptions = self._my_subscriptions[bus]
        except KeyError:
            subscriptions = dict()
        subscribers = set()
        for prefix, subscription in subscriptions.items():
            if subscription and topic.startswith(prefix):
                subscribers |= subscription
        if subscribers:
            sender = encode_peer(peer)
            json_msg = jsonapi.dumpb(
                jsonrpc.json_method(None, "pubsub.push", [sender, bus, topic, headers, message],
                                    None))
            frames = [
                zmq.Frame(""),
                zmq.Frame(""),
                zmq.Frame("RPC"),
                zmq.Frame(json_msg),
            ]
            socket = self.core().socket
            for subscriber in subscribers:
                socket.send(subscriber, flags=SNDMORE)
                socket.send_multipart(frames, copy=False)
        return len(subscribers)

    def synchronize(self):
        """Synchronize local subscriptions with the PubSubService."""
        result = next(self._results)

        subscriptions = {
            platform: {
                bus: list(subscriptions.keys())
            }
            for platform, bus_subscriptions in self._my_subscriptions.items()
            for bus, subscriptions in bus_subscriptions.items()
        }
        sync_msg = jsonapi.dumpb(dict(subscriptions=subscriptions))
        frames = ["synchronize", "connected", sync_msg]
        self.vip_socket.send_vip("", "pubsub", frames, result.ident, copy=False)

        # 2073 - python3 dictionary keys method returns a dict_keys structure that isn't serializable.
        #        added list(subscriptions.keys()) to make it like python2 list of strings.
        items = [{
            platform: {
                bus: list(subscriptions.keys())
            }
            for platform, bus_subscriptions in self._my_subscriptions.items()
            for bus, subscriptions in bus_subscriptions.items()
        }]
        for subscriptions in items:
            sync_msg = jsonapi.dumpb(dict(subscriptions=subscriptions))
            frames = ["synchronize", "connected", sync_msg]
            self.vip_socket.send_vip("", "pubsub", frames, result.ident, copy=False)

    def list(
        self,
        peer,
        prefix="",
        bus="",
        subscribed=True,
        reverse=False,
        all_platforms=False,
    ):
        """Gets list of subscriptions matching the prefix and bus for the specified peer.
        param peer: peer
        type peer: str
        param prefix: prefix of a topic
        type prefix: str
        param bus: bus
        type bus: bus
        param subscribed: subscribed or not
        type subscribed: boolean
        param reverse: reverse
        type reverse:
        :returns: List of subscriptions, i.e, list of tuples of bus, topic and
        flag to indicate if peer is a subscriber or not
        :rtype: list of tuples

        :Return Values:
        List of tuples [(topic, bus, flag to indicate if peer is a subscriber or not)]
        """
        result = next(self._results)
        list_msg = jsonapi.dumpb(
            dict(
                prefix=prefix,
                all_platforms=all_platforms,
                subscribed=subscribed,
                reverse=reverse,
                bus=bus,
            ))

        frames = ["list", list_msg]
        self.vip_socket.send_vip("", "pubsub", frames, result.ident, copy=False)
        return result

    def _add_subscription(self, prefix, callback, bus="", all_platforms=False):
        # _log.debug(f"Adding subscription prefix: {prefix} allplatforms: {all_platforms}")
        if not callable(callback):
            raise ValueError("callback %r is not callable" % (callback, ))
        try:
            if not all_platforms:
                self._my_subscriptions["internal"][bus][prefix].add(callback)
            else:
                self._my_subscriptions["all"][bus][prefix].add(callback)
                # _log.debug("SYNC: add subscriptions: {}".format(self._my_subscriptions['internal'][bus][prefix]))
        except KeyError:
            _log.error("PUBSUB something went wrong in add subscriptions")

    @dualmethod
    @spawn
    def subscribe(self,
                  peer,
                  prefix,
                  callback,
                  bus="",
                  all_platforms=False,
                  persistent_queue=None):
        """Subscribe to topic and register callback.

        Subscribes to topics beginning with prefix. If callback is
        supplied, it should be a function taking four arguments,
        callback(peer, sender, bus, topic, headers, message), where peer
        is the ZMQ identity of the bus owner sender is identity of the
        publishing peer, topic is the full message topic, headers is a
        case-insensitive dictionary (mapping) of message headers, and
        message is a possibly empty list of message parts.
        :param peer
        :type peer
        :param prefix prefix to the topic
        :type prefix str
        :param callback callback method
        :type callback method
        :param bus bus
        :type bus str
        :param platforms
        :type platforms
        :returns: Subscribe is successful or not
        :rtype: boolean

        :Return Values:
        Success or Failure
        """
        result = next(self._results)
        self._add_subscription(prefix, callback, bus, all_platforms)
        sub_msg = jsonapi.dumpb(dict(prefix=prefix, bus=bus, all_platforms=all_platforms))

        frames = ["subscribe", sub_msg]
        self.vip_socket.send_vip("", "pubsub", frames, result.ident, copy=False)
        return result

    @subscribe.classmethod
    def subscribe(cls, peer, prefix, bus="", all_platforms=False, persistent_queue=None):

        def decorate(method):
            annotate(
                method,
                set,
                "pubsub.subscriptions",
                (peer, bus, prefix, all_platforms, persistent_queue),
            )
            return method

        return decorate

    def _drop_subscription(self, prefix, callback, bus="", platform="internal"):
        """
        Drop the subscription for the specified prefix, callback and bus.
        param prefix: prefix to be removed
        type prefix: str
        param callback: callback method
        type callback: method
        param bus: bus
        type bus: bus
        return: list of topics/prefixes
        :rtype: list

        :Return Values:
        List of prefixes
        """
        topics = []
        bus_subscriptions = dict()
        if prefix is None:
            if callback is None:
                if len(self._my_subscriptions) and platform in self._my_subscriptions:
                    bus_subscriptions = self._my_subscriptions[platform]
                    if bus in bus_subscriptions:
                        topics.extend(bus_subscriptions[bus].keys())
                if not len(topics):
                    return []
            else:
                if platform in self._my_subscriptions:
                    bus_subscriptions = self._my_subscriptions[platform]
                if bus in bus_subscriptions:
                    subscriptions = bus_subscriptions[bus]
                    remove = []
                    for topic, callbacks in subscriptions.items():
                        try:
                            callbacks.remove(callback)
                        except KeyError:
                            pass
                        else:
                            topics.append(topic)
                        if not callbacks:
                            remove.append(topic)
                    for topic in remove:
                        del subscriptions[topic]
                    if not subscriptions:
                        del bus_subscriptions[bus]
                    if not bus_subscriptions:
                        del self._my_subscriptions[platform]
            if not topics:
                raise KeyError("no such subscription")
        else:
            _log.debug(f"BEFORE: {self._my_subscriptions}")
            if platform in self._my_subscriptions:
                bus_subscriptions = self._my_subscriptions[platform]
                if bus in bus_subscriptions:
                    _log.debug(f"BUS: {bus}")
                    subscriptions = bus_subscriptions[bus]
                    _log.debug(f"subscriptions: {subscriptions}")
                    if callback is None:
                        try:
                            del subscriptions[prefix]
                        except KeyError:
                            return []
                    else:
                        try:
                            callbacks = subscriptions[prefix]
                            _log.debug(f"callbacks: {callbacks}")
                        except KeyError:
                            return []
                        try:
                            callbacks.remove(callback)
                        except KeyError as e:
                            _log.debug(f"KeyError: {e}")
                            pass
                        if not callbacks:
                            try:
                                del subscriptions[prefix]
                                _log.debug(f"subscriptions: {subscriptions}")
                            except KeyError:
                                return []
                    topics = [prefix]
                    if not subscriptions:
                        del bus_subscriptions[bus]
                    if not bus_subscriptions:
                        del self._my_subscriptions[platform]
        _log.debug(f"AFTER: {self._my_subscriptions}")
        return topics

    def unsubscribe(self, peer, prefix, callback, bus="", all_platforms=False):
        """Unsubscribe and remove callback(s).

        Remove all handlers matching the given info - peer, callback and bus, which was used earlier to subscribe as
        well. If all handlers for a topic prefix are removed, the topic is also unsubscribed.
        param peer: peer
        type peer: str
        param prefix: prefix that needs to be unsubscribed
        type prefix: str
        param callback: callback method
        type callback: method
        param bus: bus
        type bus: bus
        return: success or not
        :rtype: boolean

        :Return Values:
        success or not
        """
        subscriptions = dict()
        result = next(self._results)
        if not all_platforms:
            platform = "internal"
            topics = self._drop_subscription(prefix, callback, bus, platform)
            subscriptions[platform] = dict(prefix=topics, bus=bus)
        else:
            platform = "all"
            topics = self._drop_subscription(prefix, callback, bus, platform)
            subscriptions[platform] = dict(prefix=topics, bus=bus)

        unsub_msg = jsonapi.dumpb(subscriptions)
        topics = self._drop_subscription(prefix, callback, bus)
        frames = ["unsubscribe", unsub_msg]
        self.vip_socket.send_vip("", "pubsub", frames, result.ident, copy=False)
        return result

    def publish(self, peer: str, topic: str, headers=None, message=None, bus=""):
        """Publish a message to a given topic via a peer.

        Publish headers and message to all subscribers of topic on bus.
        If peer is None, use self. Adds volttron platform version
        compatibility information to header as variables
        min_compatible_version and max_compatible version
        param peer: peer
        type peer: str
        param topic: topic for the publish message
        type topic: str
        param headers: header info for the message
        type headers: None or dict
        param message: actual message
        type message: None or any
        param bus: bus
        type bus: str
        return: Number of subscribers the message was sent to.
        :rtype: int

        :Return Values:
        Number of subscribers
        """
        if headers is None:
            headers = {}
        headers["min_compatible_version"] = min_compatible_version
        headers["max_compatible_version"] = max_compatible_version

        if peer is None:
            peer = "pubsub"

        result = next(self._results)
        args = ["publish", topic, dict(bus=bus, headers=headers, message=message)]
        self.vip_socket.send_vip("", "pubsub", args, result.ident, copy=False)
        return result

    def _check_if_protected_topic(self, topic):
        required_caps = self.protected_topics.get(topic)
        if required_caps:
            user = str(self.rpc().context.vip_message.user)
            caps = self._owner.vip.auth.get_capabilities(user)
            if not set(required_caps) <= set(caps):
                msg = ('to publish to topic "{}" requires capabilities {},'
                       " but capability list {} was"
                       " provided").format(topic, required_caps, caps)
                raise jsonrpc.exception_from_json(jsonrpc.UNAUTHORIZED, msg)

    def _handle_subsystem(self, message):
        """Handler for incoming messages
        param message: VIP message from PubSubService
        type message: dict
        """
        self._event_queue.put(message)

    @spawn
    def _process_incoming_message(self, message):
        """Process incoming messages
        param message: VIP message from PubSubService
        type message: dict
        """
        op = message.args[0]

        if op == "request_response":
            result = None
            try:
                result = self._results.pop(message.id)
            except KeyError:
                pass

            response = message.args[1]
            import struct

            if not isinstance(response, int):
                if len(response) == 4:    # integer
                    response = struct.unpack("I", response.encode("utf-8"))
                    response = response[0]
                elif len(response) == 1:    # bool
                    response = struct.unpack("?", response.encode("utf-8"))
                    response = response[0]
            if result:
                result.set(response)

        elif op == "publish":
            try:
                topic = message.args[1]
                msg = message.args[2]
            except IndexError:
                return
            try:
                headers = msg["headers"]
                message = msg["message"]
                sender = msg["sender"]
                bus = msg["bus"]
            except KeyError as exc:
                _log.error("Missing keys in pubsub message: {}".format(exc))
            else:
                self._process_callback(sender, bus, topic, headers, message)

        elif op == "list_response":
            result = None
            try:
                result = self._results.pop(message.id)
                response = message.args[1]
                if result:
                    result.set(response)
            except KeyError:
                pass
        else:
            _log.error("Unknown operation ({})".format(op))

    def _process_loop(self):
        """Incoming message processing loop"""
        for msg in self._event_queue:
            self._process_incoming_message(msg)

    def _handle_error(self, sender, message, error, **kwargs):
        """Error handler. If UnknownSubsystem error is received, it implies that agent is connected to platform that has
        OLD pubsub implementation. So messages are resent using RPC method.
        param message: Error message
        type message: dict
        param error: indicates error type
        type error: error class
        param **kwargs: variable arguments
        type **kwargs: dict
        """
        try:
            result = self._results.pop(message.id)
        except KeyError:
            return
        result.set_exception(error)


class ProtectedPubSubTopics(object):
    """Simple class to contain protected pubsub topics"""

    def __init__(self):
        self._dict = {}
        self._re_list = []

    def add(self, topic, capabilities):
        if isinstance(capabilities, str):
            capabilities = [capabilities]
        if len(topic) > 1 and topic[0] == topic[-1] == "/":
            regex = re.compile("^" + topic[1:-1] + "$")
            self._re_list.append((regex, capabilities))
        else:
            self._dict[topic] = capabilities

    def get(self, topic):
        if topic in self._dict:
            return self._dict[topic]
        for regex, capabilities in self._re_list:
            if regex.match(topic):
                return capabilities
        return None



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging

from volttron.utils.time import get_aware_utc_now, format_timestamp
from volttron.utils import jsonapi

CURRENT_STATUS = "current_status"
LAST_UPDATED = "utc_last_updated"
CONTEXT = "context"

STATUS_GOOD = "GOOD"
STATUS_BAD = "BAD"
STATUS_UNKNOWN = "UNKNOWN"
STATUS_STARTING = "STARTING"

STARTING_STATUS = STATUS_STARTING
GOOD_STATUS = STATUS_GOOD
BAD_STATUS = STATUS_BAD
UNKNOWN_STATUS = STATUS_UNKNOWN

ACCEPTABLE_STATUS = (GOOD_STATUS, BAD_STATUS, UNKNOWN_STATUS, STARTING_STATUS)

ALERT_KEY = "alert_key"

_log = logging.getLogger(__name__)


class Status(object):
    """
    The `Status` objects wraps the context status and last reported into a
    small object that can be serialized and sent across the zmq message bus.

    There are two static methods for constructing `Status` objects:
      - from_json() Expects a json string as input.
      - build() Expects at least a status in the `ACCEPTABLE_STATUS` tuple.

    The build() method also takes a context and a callback function that will
    be called when the status changes.
    """

    def __init__(self):
        self._status = GOOD_STATUS
        self._context = None
        self._last_updated = format_timestamp(get_aware_utc_now())
        self._status_changed_callback = None

    @property
    def status(self):
        return self._status

    @property
    def context(self):
        if self._context:
            if isinstance(self._context, str):
                return self._context
            return self._context.copy()
        return None

    @property
    def last_updated(self):
        return self._last_updated

    def update_status(self, status, context=None):
        """
        Updates the internal state of the `Status` object.

        This method will throw errors if the context is not serializable or
        if the status parameter is not within the ACCEPTABLE_STATUS tuple.

        :param status:
        :param context:
        :return:
        """
        if status not in ACCEPTABLE_STATUS:
            raise ValueError("Invalid status value {}".format(status))
        try:
            jsonapi.dumps(context)
        except TypeError:
            raise ValueError("Context must be JSON serializable.")

        status_changed = status != self._status
        self._status = status
        self._context = context
        self._last_updated = format_timestamp(get_aware_utc_now())

        if status_changed and self._status_changed_callback:
            self._status_changed_callback()

    def as_dict(self):
        """
        Returns a copy of the status object properties as a dictionary.

        @return:
        """
        cp = dict(status=self.status, context=self.context, last_updated=self.last_updated)
        return cp

    def as_json(self):
        """
        Serializes the object to a json string.

        Note:
            Does not serialize the change callback function.

        :return:
        """
        return jsonapi.dumps(self.as_dict())

    @staticmethod
    def from_json(data, status_changed_callback=None):
        """
        Deserializes a `Status` object and returns it to the caller.

        :param data:
        :param status_changed_callback:
        :return:
        """
        _log.debug("from_json {}".format(data))
        statusobj = Status()
        cp = jsonapi.loads(data)
        cp["_status"] = cp["status"]
        cp["_last_updated"] = cp["last_updated"]
        cp["_context"] = cp["context"]
        del cp["status"]
        del cp["last_updated"]
        del cp["context"]
        statusobj.__dict__ = cp
        statusobj._status_changed_callback = status_changed_callback
        return statusobj

    @staticmethod
    def build(status, context=None, status_changed_callback=None):
        """
        Constructs a `Status` object and initializes its state using the
        passed parameters.

        :param status:
        :param context:
        :param status_changed_callback:
        :return:
        """
        statusobj = Status()
        statusobj.update_status(status, context)
        statusobj._status_changed_callback = status_changed_callback
        return statusobj



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from volttron.client.vip.agent.subsystems.channel import Channel
from volttron.client.vip.agent.subsystems.hello import Hello
from volttron.client.vip.agent.subsystems.peerlist import PeerList
from volttron.client.vip.agent.subsystems.ping import Ping
from volttron.client.vip.agent.subsystems.pubsub import PubSub
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent.subsystems.heartbeat import Heartbeat
from volttron.client.vip.agent.subsystems.health import Health
from volttron.client.vip.agent.subsystems.configstore import ConfigStore
from volttron.client.vip.agent.subsystems.auth import Auth
# TODO Add back in with plugin architecture
# from .rmq_pubsub import RMQPubSub
from volttron.client.vip.agent.subsystems.web import WebSubSystem

__all__ = [
    "PeerList",
    "Ping",
    "RPC",
    "Hello",
    "PubSub",
    #"RMQPubSub",
    "Channel",
    "Heartbeat",
    "Health",
    "ConfigStore",
    "Auth",
    "WebSubSystem"
]
